import time

import torch
import torch.nn as nn
from loguru import logger

from llmcompressor.modifiers.utils.compression_wrapper import ModuleCompressionWrapper

try:
    import transformers
except ImportError as err:
    transformers = None
    transformers_err = err

__all__ = ["WandaWrapper"]


DEBUG = False
torch.backends.cuda.matmul.allow_tf32 = False
torch.backends.cudnn.allow_tf32 = False


class WandaWrapper(ModuleCompressionWrapper):
    """
    Runs WANDA on a single module that contains no sub-modules
    see https://arxiv.org/abs/2306.11695

    Runs SparseGPT on a single module that contains no sub-modules

    Lifecycle:
        - add_batch
        - compress
        - free

    :param name: name of module to run compression on
    :param layer: module to run compression on
    """

    def __init__(self, name, layer):
        super().__init__(name=name, layer=layer)
        self.register_buffer("scaler_row", torch.zeros(self.columns, device=self.dev))

    def add_batch(self, inp: torch.Tensor, out: torch.Tensor):
        """
        Add a batch of layer input and output data to the layer statistics calculation

        :param inp: tensor containing layer input
        :param out: tensor containing layer output
        """
        if len(inp.shape) == 2:
            inp = inp.unsqueeze(0)
        batch_size = inp.shape[0]
        if isinstance(self.layer, nn.Linear):
            if len(inp.shape) == 3:
                inp = inp.reshape((-1, inp.shape[-1]))
            inp = inp.t()

        self.scaler_row *= self.nsamples / (self.nsamples + batch_size)
        self.nsamples += batch_size
        inp = inp.type(torch.float32)
        self.scaler_row += torch.norm(inp, p=2, dim=1) ** 2 / self.nsamples

    def compress(
        self,
        sparsity: float,
        prunen: int = 0,
        prunem: int = 0,
    ):
        """
        Run pruning on the layer up to the target sparsity value.

        :param sparsity: target sparsity to reach for layer
        :param prunen: N for N:M pruning
        :param prunem: M for N:M pruning
        """
        final_shape = self.layer.weight.shape
        final_dtype = self.layer.weight.dtype
        W = self.layer.weight.data.clone()
        if isinstance(self.layer, nn.Conv2d):
            W = W.flatten(1)
        if isinstance(self.layer, transformers.Conv1D):
            W = W.t()
        W = W.float()

        tick = time.time()

        W_metric = torch.abs(W) * torch.sqrt(self.scaler_row.reshape((1, -1)))

        # initialize a mask to be all False
        W_mask = torch.zeros_like(W_metric) == 1
        if prunen != 0:
            # structured n:m sparsity
            for ii in range(W_metric.shape[1]):
                if ii % prunem == 0:
                    tmp = W_metric[:, ii : (ii + prunem)].float()
                    W_mask.scatter_(
                        1,
                        ii + torch.topk(tmp, prunen, dim=1, largest=False)[1],
                        True,
                    )
        else:
            sort_res = torch.sort(W_metric, dim=-1, stable=True)
            indices = sort_res[1][:, : int(W_metric.shape[1] * sparsity)]
            W_mask.scatter_(1, indices, True)

        W[W_mask] = 0  # set weights to zero

        logger.info("time %.2f" % (time.time() - tick))

        if isinstance(self.layer, transformers.Conv1D):
            W = W.t()

        W = W.reshape(final_shape).to(final_dtype)

        # This is a bit hacky, but FSDP updates only work if we change the weight in
        # place, clone() or direct assignment won't work
        self.layer.weight -= self.layer.weight
        self.layer.weight += W

    def free(self):
        """
        Free memory after the layer is complete
        """
        delattr(self, "scaler_row")
        super().free()



#!/usr/bin/env python
# coding=utf-8
# Copyright 2020 The HuggingFace Inc. team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Adapted from https://github.com/huggingface/transformers
# neuralmagic: no copyright

import os
from pathlib import PosixPath

from loguru import logger
from transformers import (
    AutoConfig,
    AutoTokenizer,
    DefaultDataCollator,
    HfArgumentParser,
    set_seed,
)

from llmcompressor.core import pre_initialize_structure, reset_session
from llmcompressor.pytorch.model_load.helpers import (
    fallback_to_cpu,
    get_session_model,
    initialize_recipe,
    parse_dtype,
)
from llmcompressor.recipe import Recipe, StageRunType
from llmcompressor.transformers.finetune.data.data_args import DataTrainingArguments
from llmcompressor.transformers.finetune.model_args import ModelArguments
from llmcompressor.transformers.finetune.runner import StageRunner
from llmcompressor.transformers.finetune.trainer import Trainer
from llmcompressor.transformers.finetune.training_args import TrainingArguments
from llmcompressor.transformers.sparsification.sparse_model import (
    SparseAutoModel,
    get_shared_tokenizer_src,
)
from llmcompressor.transformers.utils.helpers import detect_last_checkpoint


def train(**kwargs):
    """
    CLI entrypoint for running training
    """
    model_args, data_args, training_args = parse_args(**kwargs)
    training_args.do_train = True
    main(model_args, data_args, training_args)


def eval(**kwargs):
    """
    CLI entrypoint for running evaluation
    """
    model_args, data_args, training_args = parse_args(**kwargs)
    training_args.do_eval = True
    main(model_args, data_args, training_args)


def oneshot(**kwargs):
    """
    CLI entrypoint for running oneshot calibration
    """
    model_args, data_args, training_args = parse_args(**kwargs)
    training_args.do_oneshot = True
    main(model_args, data_args, training_args)


# alias
one_shot = oneshot


def apply(**kwargs):
    """
    CLI entrypoint for any of training, eval, predict or oneshot
    """
    report_to = kwargs.get("report_to", None)
    model_args, data_args, training_args = parse_args(**kwargs)
    training_args.run_stages = True
    if report_to is None:  # user didn't specify any reporters
        # get rid of the reporters inferred from hugging face
        training_args.report_to = []
    main(model_args, data_args, training_args)


def compress(**kwargs):
    apply(**kwargs)


def load_dataset(dataset_name: str, **kwargs):
    parser = HfArgumentParser(
        (ModelArguments, DataTrainingArguments, TrainingArguments)
    )
    model_args, data_args, training_args = parser.parse_dict(kwargs)
    data_args["dataset_name"] = dataset_name


def parse_args(**kwargs):
    """
    Parses kwargs by grouping into model, data or training arg groups:
        * model_args in src/llmcompressor/transformers/finetune/model_args.py
        * data_args in src/llmcompressor/transformers/finetune/data/data_args.py
        * training_args in src/llmcompressor/transformers/finetune/training_args.py
    """
    parser = HfArgumentParser(
        (ModelArguments, DataTrainingArguments, TrainingArguments)
    )
    if not kwargs:
        model_args, data_args, training_args = parser.parse_args_into_dataclasses()
    else:
        model_args, data_args, training_args = parser.parse_dict(kwargs)

    if training_args.recipe_args is not None:
        if not isinstance(training_args.recipe_args, dict):
            arg_dict = {}
            for recipe_arg in training_args.recipe_args:
                key, value = recipe_arg.split("=")
                arg_dict[key] = value
            training_args.recipe_args = arg_dict

    return model_args, data_args, training_args


def initialize_model_from_path(
    model_args: ModelArguments,
    training_args: TrainingArguments,
):
    last_checkpoint = detect_last_checkpoint(training_args, model_args=model_args)
    # Load pretrained model
    # The .from_pretrained methods guarantee that only one local process can
    # concurrently download model & vocab.
    model_path = model_args.model
    config = AutoConfig.from_pretrained(
        model_args.config_name if model_args.config_name else model_path,
        cache_dir=model_args.cache_dir,
        revision=model_args.model_revision,
        use_auth_token=True if model_args.use_auth_token else None,
        tie_word_embeddings=model_args.tie_word_embeddings,
        trust_remote_code=model_args.trust_remote_code_model,
    )
    teacher_config = (
        AutoConfig.from_pretrained(
            model_args.distill_teacher,
            use_auth_token=True if model_args.use_auth_token else None,
            tie_word_embeddings=model_args.tie_word_embeddings,
            trust_remote_code=model_args.trust_remote_code_model,
        )
        if model_args.distill_teacher
        else None
    )

    model_path = (
        last_checkpoint or model_args.model
        if hasattr(model_args, "model")
        else model_args.model_name_or_path
    )

    # Set seed before initializing model.
    set_seed(training_args.seed)

    # Fallback to CPU if GPU requested and not available
    training_args.oneshot_device = fallback_to_cpu(training_args.oneshot_device)

    # Trainer handles device assignment for FSDP and training, don't do mapping here
    # if running oneshot outside of FSDP, apply user device settings
    device_map = None
    fsdp_enabled = os.environ.get("ACCELERATE_USE_FSDP", "false") == "true"
    if not fsdp_enabled and training_args.do_oneshot:
        device_map = training_args.oneshot_device
        logger.warning(f"Moving {model_path} to device {device_map} for One-Shot")
    elif not fsdp_enabled:
        device_map = "auto"
    model_kwargs = {
        "config": config,
        "cache_dir": model_args.cache_dir,
        "revision": model_args.model_revision,
        "use_auth_token": True if model_args.use_auth_token else None,
        "torch_dtype": parse_dtype(model_args.precision),
        "device_map": device_map,
        "trust_remote_code": model_args.trust_remote_code_model,
    }
    teacher_device_map = None if fsdp_enabled else "auto"
    teacher_kwargs = {
        "config": teacher_config,
        "cache_dir": model_args.cache_dir,
        "use_auth_token": True if model_args.use_auth_token else None,
        "torch_dtype": parse_dtype(model_args.precision),
        "device_map": teacher_device_map,
        "trust_remote_code": model_args.trust_remote_code_model,
    }
    # this calls from_pretrained under the hood so should be FSDP safe
    model = SparseAutoModel.text_generation_from_pretrained(
        model_name_or_path=model_path,
        sequence_length=None,  # use model default
        **model_kwargs,
    )

    teacher = (
        SparseAutoModel.text_generation_from_pretrained(
            model_name_or_path=model_args.distill_teacher,
            sequence_length=None,  # use model default
            **teacher_kwargs,
        )
        if model_args.distill_teacher is not None
        else None
    )

    return teacher, model_path, model


def initialize_tokenizer_from_path(model_args, model, teacher):
    tokenizer_src = model_args.tokenizer
    tokenizer_src = tokenizer_src or get_shared_tokenizer_src(model, teacher)
    tokenizer = AutoTokenizer.from_pretrained(
        tokenizer_src,
        cache_dir=model_args.cache_dir,
        use_fast=True,
        revision=model_args.model_revision,
        use_auth_token=True if model_args.use_auth_token else None,
        trust_remote_code=model_args.trust_remote_code_model,
    )

    return tokenizer


def main(
    model_args: ModelArguments,
    data_args: DataTrainingArguments,
    training_args: TrainingArguments,
):
    """
    Main entrypoint for finetuning text generation models. A model can be loaded from
    Hugging Face or disk, and resuming training from a checkpoint is supported.

    Lifecycle:
        - SparseAutoModel.text_generation_from_pretrained if model provided as
            string for model and teacher
        - AutoTokenizer.from_pretrained() if tokenizer provided as
            string for tokenizer
        - StageRunner.populate_datasets()
        - Trainer()
            - SessionMixIn()
            - HFTransformersTrainer()
        - StageRunner.train() and/or evaluate() and/or predict() and/or oneshot()

    :param model_args: Arguments pertaining to which model/config/tokenizer we are
    going to fine-tune from
    :param data_args: Arguments pertaining to what data we are going to input our model
    for training and eval
    :param training_args: Arguments pertaining to training loop configuration
    """

    # Temporary warning, to be removed
    if model_args.tie_word_embeddings is True:
        logger.warning(
            "The tie_word_embeddings flag is by default set to False. "
            "This guarantees that the one-shot algorithm saves the final "
            "weights without errors. Detected tie_word_embeddings=True. "
            "This may cause issues with the one-shot algorithm on save. "
        )

    # Setup based on stage types if running stage mode
    if training_args.run_stages and training_args.recipe is not None:
        recipe_obj = Recipe.create_instance(training_args.recipe)
        for stage in recipe_obj.stages:
            run_type = stage.infer_run_type()
            if run_type is StageRunType.ONESHOT:
                training_args.do_oneshot = True
            elif run_type is StageRunType.TRAIN:
                training_args.do_train = True

    # Summary on each process
    logger.warning(
        f"Process rank: {training_args.local_rank}, device: {training_args.device}, "
        f"n_gpu: {training_args.n_gpu}, "
        f"distributed training: {bool(training_args.local_rank != -1)}, "
        f"16-bits training: {training_args.fp16}"
    )
    logger.info(f"Training/evaluation parameters {training_args}")

    # Detecting last checkpoint.
    last_checkpoint = None
    teacher = model_args.distill_teacher
    model_path = None
    model = model_args.model
    # Load tokenizer
    # distill TODO: support for different tokenizer for teacher?
    tokenizer = model_args.tokenizer

    if isinstance(model, str) or isinstance(model, PosixPath):
        (teacher, model_path, model) = initialize_model_from_path(
            model_args,
            training_args,
        )

    if teacher is not None:
        teacher.eval()

    if isinstance(tokenizer, str) or tokenizer is None:
        tokenizer = initialize_tokenizer_from_path(model_args, model, teacher)

    pre_initialize_structure(model=model)

    # initialize session manager
    initialize_recipe(model, None)

    # Load datasets
    stage_runner = StageRunner(
        model_args=model_args, data_args=data_args, training_args=training_args
    )
    add_labels = training_args.do_train or training_args.run_stages
    stage_runner.populate_datasets(tokenizer=tokenizer, add_labels=add_labels)
    train_dataset = stage_runner.get_dataset_split("train")
    eval_dataset = stage_runner.get_dataset_split("validation")
    calib_dataset = stage_runner.get_dataset_split("calibration")

    # Initialize our Trainer
    data_collator = DefaultDataCollator()
    trainer = Trainer(
        model_init=get_session_model,
        teacher=teacher,
        recipe=training_args.recipe,
        recipe_args=training_args.recipe_args,
        args=training_args,
        data_args=data_args,
        train_dataset=train_dataset or calib_dataset,
        eval_dataset=eval_dataset,
        tokenizer=tokenizer,
        data_collator=data_collator,
    )
    stage_runner.trainer = trainer

    # alternating Training/One-shot
    if training_args.run_stages:
        checkpoint = None
        if last_checkpoint is not None:
            checkpoint = last_checkpoint
        stage_runner.run_sequential_stages(checkpoint)

        # exit immediately
        return

    # Training
    if training_args.do_train:
        checkpoint = None
        if training_args.resume_from_checkpoint is not None:
            checkpoint = training_args.resume_from_checkpoint
        elif last_checkpoint is not None:
            checkpoint = last_checkpoint
        stage_runner.train(checkpoint)

    # One Shot
    if training_args.do_oneshot:
        stage_runner.one_shot()

    # Evaluation
    if training_args.do_eval:
        stage_runner.evaluate()

    # Prediction
    if training_args.do_predict:
        stage_runner.predict()

    # Clean up the CompressionSession before exit if requested
    if training_args.clear_sparse_session:
        reset_session()


if __name__ == "__main__":
    apply()



import json
import os
from typing import Any, Dict, List, Optional

import torch
from loguru import logger
from safetensors import safe_open
from torch.nn import Module

from llmcompressor.core import active_session, create_session, pre_initialize_structure
from llmcompressor.pytorch.utils import ModuleSparsificationInfo

COMPLETED_STAGES_FILENAME = "completed_stages.json"

__all__ = [
    "log_model_load",
    "initialize_recipe",
    "save_model_and_recipe",
    "fallback_to_cpu",
    "parse_dtype",
    "get_session_model",
    "get_completed_stages",
    "save_completed_stages",
]

RECIPE_FILE_NAME = "recipe.yaml"


def log_model_load(
    model: Module, model_name_or_path: str, model_type: str, delayed_load: bool
):
    """
    Log the state of a loaded model including sparsity and
    prunable params information.

    :param model: the loaded model
    :param model_name_or_path: the original name of or path to the model that loaded
    :param model_type: specify the type of model loaded for logging;
        ex one of [model, student, teacher]
    :param delayed_load: True if this model load was delayed until after
        recipe instantiation due to QAT or other architectural state changes
    """
    if delayed_load:
        logger.info(
            f"Delayed load of model {model_name_or_path} detected. "
            f"Will print out model information once LLMCompressor recipes have loaded"
        )
        return

    sparsification_info = ModuleSparsificationInfo(model)

    logger.info(
        f"Loaded {model_type} from {model_name_or_path} "
        f"with {sparsification_info.params_total} total params. "
        f"Of those there are {sparsification_info.params_prunable_total} prunable "
        f"params which have {sparsification_info.params_prunable_sparse_percent} "
        "avg sparsity."
    )
    model_type = (
        "sparse" if sparsification_info.params_prunable_sparse_percent > 5 else "dense"
    )
    logger.info(
        f"{model_type} model detected, "
        f"all sparsification info: {sparsification_info}"
    )


def initialize_recipe(model: Module, recipe_path: str):
    """
    Initializes a recipe that has been previously applied to the model

    :param model: PyTorch model to apply structure to
    :param recipe_path: path to recipe to apply to the model
    """
    if not active_session():
        create_session()
    pre_initialize_structure(model=model, recipe=recipe_path)

    # no need to reload if no recipe was applied
    if recipe_path is None:
        return

    session = active_session()
    num_stages = len(session.lifecycle.recipe_container.compiled_recipe.stages)
    msg = (
        "an unstaged recipe"
        if num_stages == 1
        else f"a staged recipe with {num_stages} stages"
    )
    logger.info(f"Applied {msg} to the model")


def save_model_and_recipe(
    model: Module,
    save_path: str,
    tokenizer: Optional[Any] = None,
    save_safetensors: bool = False,
    save_compressed: bool = False,
):
    """
    Save a model, tokenizer and the currently loaded recipe to file

    :param model: pytorch model to save
    :param save_path: path to save output to
    :param tokenizer: model tokenizer to save
    :param save_safetensors: whether to save as safetensors or pickle (bin)
    :param save_compressed: whether to compress sparse weights on disk
    """

    model.save_pretrained(
        save_path, save_compressed=save_compressed, safe_serialization=save_safetensors
    )

    if tokenizer is not None:
        tokenizer.save_pretrained(save_path)

    logger.info("Saving output to {}".format(os.path.abspath(save_path)))

    recipe_path = os.path.join(save_path, RECIPE_FILE_NAME)
    session = active_session()
    recipe_yaml_str = session.get_serialized_recipe()
    with open(recipe_path, "w") as fp:
        fp.write(recipe_yaml_str)

    # copy python files from cache dir to save_path if any
    _copy_python_files_from_model_cache(model, save_path)


def fallback_to_cpu(device: str) -> str:
    """
    Takes in a device string and forces it to cpu if cuda is not available

    :param device: device id to check
    :return: device modified for CUDA status
    """
    if "cuda" in device and not torch.cuda.is_available():
        logger.warning(
            f"Requested {device} but CUDA is not available, falling back to CPU"
        )
        return "cpu"

    return device


def parse_dtype(dtype_arg: str) -> torch.dtype:
    """
    :param dtype_arg: dtype string to parse
    :return: torch.dtype parsed from input string
    """
    dtype = "auto"  # get precision from model by default
    if dtype_arg == "half" or dtype_arg == "float16":
        dtype = torch.float16
    elif dtype_arg == "bfloat16":
        dtype = torch.bfloat16
    elif dtype_arg == "full" or dtype_arg == "float32":
        dtype = torch.float32

    return dtype


def get_session_model() -> Optional[Module]:
    """
    :return: pytorch module stored by the active CompressionSession,
        or None if no session is active
    """
    session = active_session()
    if not session:
        return None

    active_model = session.state.model
    return active_model


def get_completed_stages(checkpoint_dir: Any) -> List[str]:
    """
    Given a checkpoint directory for a staged run, get the list of stages that
    have completed in a prior run if the checkpoint_dir is a string

    :param checkpoint_dir: path to staged checkpoint
    :return: list of completed stage names
    """
    if isinstance(checkpoint_dir, str):
        stage_path = os.path.join(checkpoint_dir, COMPLETED_STAGES_FILENAME)
        if os.path.exists(stage_path):
            with open(stage_path) as stage_file:
                stage_data = json.load(stage_file)
                return stage_data["completed"]

    return []


def save_completed_stages(checkpoint_dir: str, completed_stages: List[str]):
    """
    Save a list of completed stages to a checkpoint directory

    :param checkpoint_dir: model checkpoint directory to save stages to
    :param completed_stages: list of stage names that have been run
    """
    stage_path = os.path.join(checkpoint_dir, COMPLETED_STAGES_FILENAME)
    with open(stage_path, "w") as out_file:
        json.dump({"completed": completed_stages}, out_file)


def load_safetensors_state_dict(file_path: str) -> Dict[str, torch.Tensor]:
    """
    Load a safetensors file from disk

    :param file_path: path to the safetensors file
    :return: dictionary of safetensors data
    """
    with safe_open(file_path, framework="pt", device="cpu") as f:
        return {key: f.get_tensor(key) for key in f.keys()}


def _copy_python_files_from_model_cache(model: Module, save_path: str):
    config = model.config
    cache_dir = None
    if hasattr(config, "_name_or_path"):
        import os
        import shutil

        cache_dir = config._name_or_path
        for file in os.listdir(cache_dir):
            full_file_name = os.path.join(cache_dir, file)
            if file.endswith(".py") and os.path.isfile(full_file_name):
                logger.debug(f"Transferring {full_file_name} to {save_path}")
                shutil.copy(full_file_name, save_path)



# Copyright (c) 2021 - present / Neuralmagic, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from copy import deepcopy
from typing import Optional

from llmcompressor.transformers.finetune.data import TextGenerationDataset


@TextGenerationDataset.register(name="ultrachat_200k")
class UltraChatDataset(TextGenerationDataset):
    """
    Child text generation class for the Ultra Chat 200k dataset

    :param data_args: configuration settings for dataset loading
    :param split: split from dataset to load, for instance `test` or `train[:5%]`
    :param tokenizer: tokenizer to use on dataset
    """

    DEFAULT_CHAT_TEMPLATE = (
        "{% for message in messages %}\n"
        "{% if message['role'] == 'user' %}\n"
        "{{ '<|user|>\n' + message['content'] + eos_token }}\n"
        "{% elif message['role'] == 'system' %}\n"
        "{{ '<|system|>\n' + message['content'] + eos_token }}\n"
        "{% elif message['role'] == 'assistant' %}\n"
        "{{ '<|assistant|>\n'  + message['content'] + eos_token }}\n"
        "{% endif %}\n"
        "{% if loop.last and add_generation_prompt %}\n"
        "{{ '<|assistant|>' }}\n{% endif %}\n{% endfor %}"
    )

    def __init__(self, data_args, split, tokenizer):
        data_args = deepcopy(data_args)
        data_args.dataset = "HuggingFaceH4/ultrachat_200k"

        if split in ["train", "test"]:
            split += "_sft"

        super().__init__(
            text_column="messages",
            data_args=data_args,
            split=split,
            tokenizer=tokenizer,
        )

        if (
            not hasattr(self.tokenizer, "chat_template")
            or self.tokenizer.chat_template is None
        ):
            self.tokenizer.chat_template = self.DEFAULT_CHAT_TEMPLATE

    def get_raw_dataset(self, cache_dir: Optional[str] = None):
        """
        Load the raw dataset from Hugging Face, using cached copy if available.
        Additionally reformats the entries to fit the alpaca template.

        :param cache_dir: disk location to search for cached dataset
        :return: the requested dataset
        """
        raw_dataset = super().get_raw_dataset(cache_dir=cache_dir)

        # helper fn for restructuring each dataset entry using the chat template
        def restructure_fn(sample):
            if sample["messages"][0]["role"] != "system":
                sample["messages"].insert(0, {"role": "system", "content": ""})

            sample["messages"] = self.tokenizer.apply_chat_template(
                sample["messages"], tokenize=False, add_generation_prompt=False
            )
            return sample

        raw_dataset = self.map(
            raw_dataset,
            function=restructure_fn,
            batched=False,
            remove_columns=[],
            num_proc=self.data_args.preprocessing_num_workers,
            load_from_cache_file=not self.data_args.overwrite_cache,
            desc="Restructuring Ultra Chat Dataset",
        )
        return raw_dataset



from typing import List, Tuple

from loguru import logger
from torch.nn import Module


def get_GPU_memory_usage() -> List[Tuple]:
    try:
        import pynvml
        from pynvml import NVMLError

        try:
            pynvml.nvmlInit()
        except NVMLError as _err:
            logger.warning(f"Pynml library error:\n {_err}")
            return []

        device_count = pynvml.nvmlDeviceGetCount()
        usage = []  # [(percentage, total_memory_MB)]

        # Iterate through all GPUs
        for i in range(device_count):
            handle = pynvml.nvmlDeviceGetHandleByIndex(i)
            mem_info = pynvml.nvmlDeviceGetMemoryInfo(handle)
            memory_usage_percentage = mem_info.used / mem_info.total
            total_memory_gb = mem_info.total / (1024**3)
            usage.append(
                (memory_usage_percentage, total_memory_gb),
            )
        pynvml.nvmlShutdown()
        return usage

    except ImportError:
        logger.warning("Failed to obtain GPU usage from pynvml")
        return []


def get_layer_size_bytes(module: Module) -> float:
    param_size = 0
    buffer_size = 0

    for param in module.parameters():
        param_size += param.nelement() * param.element_size()

    for buffer in module.buffers():
        buffer_size += buffer.nelement() * buffer.element_size()

    total_size = param_size + buffer_size
    total_size_mb = total_size / (1024**2)  # Convert bytes to MB

    return total_size_mb



from dataclasses import dataclass, field
from typing import List, Optional

from transformers import TrainingArguments as HFTrainingArgs

__all__ = ["TrainingArguments"]


@dataclass
class TrainingArguments(HFTrainingArgs):
    """
    Training arguments specific to LLM Compressor Transformers workflow

    :param best_model_after_epoch (`int`, *optional*, defaults to None):
        The epoch after which best model will be saved; used in conjunction
        with `load_best_model_at_end` and `metric_for_best_model` training
        arguments
    """

    recipe: Optional[str] = field(
        default=None,
        metadata={
            "help": "Path to a LLM Compressor sparsification recipe",
        },
    )
    recipe_args: Optional[List[str]] = field(
        default=None,
        metadata={
            "help": (
                "List of recipe arguments to evaluate, of the format key1=value1 "
                "key2=value2"
            )
        },
    )
    save_compressed: Optional[bool] = field(
        default=True,
        metadata={"help": "Whether to compress sparse models during save"},
    )
    do_oneshot: Optional[bool] = field(
        default=False,
        metadata={"help": "Whether to run one-shot calibration"},
    )
    run_stages: Optional[bool] = field(
        default=False, metadata={"help": "Whether to trigger recipe stage by stage"}
    )
    oneshot_device: Optional[str] = field(
        default="cuda:0",
        metadata={"help": "Device to run oneshot calibration on"},
    )
    clear_sparse_session: Optional[bool] = field(
        default=False,
        metadata={"help": "Whether to clear CompressionSession data between runs."},
    )
    save_safetensors: Optional[bool] = field(
        default=True,
        metadata={
            "help": "Use safetensors saving and loading for state dicts instead of "
            "default torch.load and torch.save."
        },
    )
    output_dir: str = field(
        default="./output",
        metadata={
            "help": "The output directory where the model predictions and "
            "checkpoints will be written."
        },
    )

    @property
    def place_model_on_device(self):
        return False



"""
Utility / helper functions
"""

import os
import random
import re
from collections import OrderedDict, namedtuple
from contextlib import contextmanager
from copy import deepcopy
from typing import Any, Dict, Iterable, List, Mapping, Optional, Tuple, Union

import numpy
import torch
from loguru import logger
from packaging import version
from torch import Tensor
from torch.nn import Embedding, Linear, Module, Parameter
from torch.nn.modules.conv import Conv2d, Conv3d, _ConvNd
from torch.optim.optimizer import Optimizer
from torch.utils.data import DataLoader

try:
    quant_err = None
    from torch.nn.qat import Conv2d as QATConv2d
    from torch.nn.qat import Linear as QATLinear
    from torch.quantization import QuantWrapper
except Exception as _err:
    quant_err = _err
    QuantWrapper = None
    QATLinear = None
    QATConv2d = None

from llmcompressor.utils import create_dirs, save_numpy

try:
    from torch.nn.qat import Conv3d as QATConv3d
except Exception as _err:
    quant_conv3d_err = _err
    QATConv3d = None


try:
    from transformers.modeling_utils import Conv1D as GPTConv1D
except Exception as _err:
    gpt_conv1d_err = _err
    GPTConv1D = None


__all__ = [
    "default_device",
    "device_of",
    "get_optim_learning_rate",
    "get_optim_groups_learning_rates",
    "set_optim_learning_rate",
    "early_stop_data_loader",
    "infinite_data_loader",
    "tensors_batch_size",
    "tensors_to_device",
    "tensors_to_precision",
    "tensors_module_forward",
    "tensor_export",
    "tensors_export",
    "tensor_density",
    "tensor_sparsity",
    "tensor_list_sparsity",
    "tensor_sample",
    "mask_difference",
    "get_layer",
    "get_terminal_layers",
    "get_conv_layers",
    "get_linear_layers",
    "get_prunable_layers",
    "get_quantizable_layers",
    "swap_modules",
    "get_named_layers_and_params_by_regex",
    "any_str_or_regex_matches_param_name",
    "NamedLayerParam",
    "get_layer_param",
    "set_deterministic_seeds",
    "torch_distributed_zero_first",
    "thin_model_from_checkpoint",
    "MEMORY_BOUNDED",
    "memory_aware_threshold",
    "detach",
    "adjust_quantization_for_onnx_export",
    "get_dependency_order",
]


_PARSED_TORCH_VERSION = version.parse(torch.__version__)


##############################
#
# pytorch device helpers
#
##############################


def default_device() -> str:
    """
    :return: the device that should be defaulted to for the current setup.
        if multiple gpus are available then will return a string with all of them,
        else if single gpu available then will return cuda,
        else returns cpu
    """

    if not torch.cuda.is_available():
        return "cpu"

    if torch.cuda.device_count() < 2:
        return "cuda"

    device_ids = [str(i) for i in range(torch.cuda.device_count())]

    return "cuda:{}".format(",".join(device_ids))


def device_of(inputs: Any):
    if isinstance(inputs, Tensor):
        return inputs.device
    elif isinstance(inputs, Mapping):
        for tens in inputs.values():
            return device_of(tens)
    elif isinstance(inputs, Iterable):
        return device_of(inputs[0])
    else:
        raise RuntimeError("Unknown type of inputs to device_of function")
    return default_device()


##############################
#
# pytorch optim helpers
#
##############################


def get_optim_learning_rate(optim: Optimizer) -> float:
    """
    :param optim: The optimizer to get the learning rate for

    :return: convenience function to get the first learning rate for any of
        the param groups in the optimizer
    """
    for param_group in optim.param_groups:
        return param_group["lr"]

    raise RuntimeError("cannot get learning_rate, no param_groups available")


def get_optim_groups_learning_rates(optim: Optimizer) -> List[float]:
    """
    :param optim: The optimizer to get the learning rates for

    :return: get a list of tuples corresponding to the learning rates for the
        param groups in the optimizer
    """
    return [group["lr"] for group in optim.param_groups]


def set_optim_learning_rate(
    optim: Optimizer, value: float, groups: Optional[List[int]] = None
):
    """
    :param optim: The optimizer to set the learning rate for
    :param value: the learning rate to set for the optimizer,
        will set all param groups in the optim to this value
    """
    for index, group in enumerate(optim.param_groups):
        if not groups or index in groups:
            group["lr"] = value


##############################
#
# pytorch data loader helpers
#
##############################


def early_stop_data_loader(data_loader: DataLoader, early_stop_steps: int):
    """
    An iterator that goes through the data_loader for yields and stops
    after early_stop_steps instead of the full loader

    :param data_loader: the data loader to continually repeat
    :param early_stop_steps: if set, the number of steps to run and break out early
        instead of running all of the steps in the data loader,
        if < 1 then will run the full length
    :return: an iterable for the never ending data loader
    """
    counter = 0

    for data in data_loader:
        yield data
        counter += 1

        if 0 < early_stop_steps <= counter:
            break


def infinite_data_loader(
    data_loader: DataLoader, early_stop_steps: int = -1, cache: bool = False
):
    """
    A never ending data loader that will keep repeating the one passed in.
    Will additionally cache the data if requested.

    :param data_loader: the data loader to continually repeat
    :param early_stop_steps: if set, the number of steps to run and break out early
        instead of running all of the steps in the data loader
    :param cache: True to cache the results in memory and return those on
        subsequent requests, False otherwise
    :return: an iterable for the never ending data loader
    """
    cached = None

    while True:
        if not cache or cached is None:
            cached = []

            for data in early_stop_data_loader(data_loader, early_stop_steps):
                if cache:
                    cached.append(deepcopy(data))

                yield data
        else:
            for data in cached:
                yield data


##############################
#
# pytorch tensor helper functions
#
##############################


NamedLayerParam = namedtuple(
    "NamedLayerParam", ["layer_name", "layer", "param_name", "param"]
)


def tensors_batch_size(tensors: Union[Tensor, Iterable[Tensor], Dict[Any, Tensor]]):
    """
    Default function for getting the batch size from a tensor or collection of tensors.
    Returns the batch size (zeroth index for shape) of the first found tensor.

    Supported use cases:
        - single tensor
        - Dictionary of single tensors
        - Dictionary of iterable of tensors
        - Dictionary of dictionary of tensors
        - Iterable of single tensors
        - Iterable of iterable of tensors
        - Iterable of dictionary of tensors

    :param tensors: the tensor or collection of tensors to get a batch size from,
        taken from the first found tensor
    :return: the batch size (0th element of shape) of the first contained
        tensor in the data
    """
    if isinstance(tensors, Tensor):
        return tensors.shape[0]

    if isinstance(tensors, Dict):
        for key, tens in tensors.items():
            batch_size = tensors_batch_size(tens)

            if batch_size > -1:
                return batch_size

    if isinstance(tensors, Iterable):
        for tens in tensors:
            batch_size = tensors_batch_size(tens)

            if batch_size > -1:
                return batch_size

    return -1


def tensors_to_device(
    tensors: Union[Tensor, Iterable[Tensor], Dict[Any, Tensor]], device: str
) -> Union[Tensor, Iterable[Tensor], Dict[Any, Tensor]]:
    """
    Default function for putting a tensor or collection of tensors to the proper device.
    Returns the tensor references after being placed on the proper device.

    Supported use cases:
        - single tensor
        - Dictionary of single tensors
        - Dictionary of iterable of tensors
        - Dictionary of dictionary of tensors
        - Iterable of single tensors
        - Iterable of iterable of tensors
        - Iterable of dictionary of tensors

    :param tensors: the tensors or collection of tensors to put onto a device
    :param device: the string representing the device to put the tensors on,
        ex: 'cpu', 'cuda', 'cuda:1'
    :return: the tensors or collection of tensors after being placed on the device
    """
    if isinstance(tensors, Tensor):
        return tensors.to(device)

    if isinstance(tensors, OrderedDict):
        return OrderedDict(
            [(key, tensors_to_device(tens, device)) for key, tens in tensors.items()]
        )

    if isinstance(tensors, Mapping):
        return {key: tensors_to_device(tens, device) for key, tens in tensors.items()}

    if isinstance(tensors, tuple):
        return tuple(tensors_to_device(tens, device) for tens in tensors)

    if isinstance(tensors, Iterable):
        return [tensors_to_device(tens, device) for tens in tensors]

    raise ValueError(
        "unrecognized type for tensors given of {}".format(tensors.__class__.__name__)
    )


def tensors_to_precision(
    tensors: Union[Tensor, Iterable[Tensor], Dict[Any, Tensor]], full_precision: bool
) -> Union[Tensor, Iterable[Tensor], Dict[Any, Tensor]]:
    """
    :param tensors: the tensors to change the precision of
    :param full_precision: True for full precision (float 32) and
        False for half (float 16)
    :return: the tensors converted to the desired precision
    """
    if isinstance(tensors, Tensor):
        return tensors.float() if full_precision else tensors.half()

    if isinstance(tensors, Mapping):
        return {
            key: tensors_to_precision(tens, full_precision)
            for key, tens in tensors.items()
        }

    if isinstance(tensors, tuple):
        return tuple(tensors_to_precision(tens, full_precision) for tens in tensors)

    if isinstance(tensors, Iterable):
        return [tensors_to_precision(tens, full_precision) for tens in tensors]

    raise ValueError(
        "unrecognized type for tensors given of {}".format(tensors.__class__.__name__)
    )


def tensors_module_forward(
    tensors: Union[Tensor, Iterable[Tensor], Mapping[Any, Tensor]],
    module: Module,
    check_feat_lab_inp: bool = True,
) -> Any:
    """
    Default function for calling into a model with data for a forward execution.
    Returns the model result.
    Note, if an iterable the features to be passed into the model are considered
    to be at index 0 and other indices are for labels.

    Supported use cases: single tensor,
    iterable with first tensor taken as the features to pass into the model

    :param tensors: the data to be passed into the model, if an iterable the features
        to be passed into the model are considered to be at index 0 and other indices
        are for labels
    :param module: the module to pass the data into
    :param check_feat_lab_inp: True to check if the incoming tensors looks like
        it's made up of features and labels ie a tuple or list with 2 items
        (typical output from a data loader) and will call into the model with just
        the first element assuming it's the features False to not check
    :return: the result of calling into the model for a forward pass
    """
    if (
        (isinstance(tensors, tuple) or isinstance(tensors, List))
        and len(tensors) == 2
        and check_feat_lab_inp
    ):
        # assume if this is a list or tuple of 2 items that it is made up of
        # (features, labels) pass the features into a recursive call for the model
        return tensors_module_forward(tensors[0], module, check_feat_lab_inp=False)

    if isinstance(tensors, Tensor):
        return module(tensors)

    if isinstance(tensors, Mapping):
        return module(**tensors)

    if isinstance(tensors, Iterable):
        return module(*tensors)

    raise ValueError(
        "unrecognized type for data given of {}".format(tensors.__class__.__name__)
    )


def tensor_export(
    tensor: Union[Tensor, Dict[str, Tensor], Iterable[Tensor]],
    export_dir: str,
    name: str,
    npz: bool = True,
) -> str:
    """
    :param tensor: tensor to export to a saved numpy array file
    :param export_dir: the directory to export the file in
    :param name: the name of the file, .npy will be appended to it
    :param npz: True to export as an npz file, False otherwise
    :return: the path of the numpy file the tensor was exported to
    """
    if isinstance(tensor, Tensor):
        tensor = tensor.detach().cpu().numpy()
    elif isinstance(tensor, Dict):
        tensor = OrderedDict(
            (key, val.detach().cpu().numpy()) for key, val in tensor.items()
        )
    elif isinstance(tensor, Iterable):
        tensor = [
            val.detach().cpu().numpy() if isinstance(val, Tensor) else val
            for val in tensor
        ]
    else:
        raise ValueError("Unrecognized type given for tensorr {}".format(tensor))

    return save_numpy(tensor, export_dir, name, npz)


def tensors_export(
    tensors: Union[Tensor, Iterable[Tensor]],
    export_dir: str,
    name_prefix: str,
    counter: int = 0,
    break_batch: bool = False,
) -> List[str]:
    """
    :param tensors: the tensors to export to a saved numpy array file
    :param export_dir: the directory to export the files in
    :param name_prefix: the prefix name for the tensors to save as, will append
        info about the position of the tensor in a list or dict in addition
        to the .npy file format
    :param counter: the current counter to save the tensor at
    :param break_batch: treat the tensor as a batch and break apart into
        multiple tensors
    :return: the exported paths
    """
    create_dirs(export_dir)
    exported_paths = []
    if break_batch:
        _tensors_export_batch(tensors, export_dir, name_prefix, counter, exported_paths)
    else:
        _tensors_export_recursive(
            tensors, export_dir, name_prefix, counter, exported_paths
        )

    return exported_paths


def _tensors_export_recursive(
    tensors: Union[Tensor, Iterable[Tensor]],
    export_dir: str,
    name_prefix: str,
    counter: int,
    exported_paths: List[str],
):
    if isinstance(tensors, Tensor):
        exported_paths.append(
            tensor_export(tensors, export_dir, "{}-{:04d}".format(name_prefix, counter))
        )

        return

    if isinstance(tensors, Iterable):
        for index, tens in enumerate(tensors):
            _tensors_export_recursive(
                tens,
                export_dir,
                name_prefix,
                counter + index,
                exported_paths,
            )

        return

    raise ValueError(
        "unrecognized type for tensors given of {}".format(tensors.__class__.__name__)
    )


def _tensors_export_batch(
    tensors: Union[Tensor, Iterable[Tensor]],
    export_dir: str,
    name_prefix: str,
    counter: int,
    exported_paths: List[str],
):
    if isinstance(tensors, Tensor):
        if len(tensors.shape) == 1:
            exported_paths.append(
                tensor_export(
                    tensors, export_dir, "{}-{:04d}".format(name_prefix, counter)
                )
            )
            return

        for index, tens in enumerate(tensors):
            exported_paths.append(
                tensor_export(
                    tens, export_dir, "{}-{:04d}".format(name_prefix, counter + index)
                )
            )

        return

    if isinstance(tensors, Iterable):
        # TODO: I am breaking something here? - dbogunowicz
        for index, tens in enumerate(tensors):
            exported_paths.append(
                tensor_export(
                    tens, export_dir, "{}-{:04d}".format(name_prefix, counter + index)
                )
            )

        return

    raise ValueError(
        "unrecognized type for tensors given of {}".format(tensors.__class__.__name__)
    )


def tensor_sparsity(
    tens: Tensor, dim: Union[None, int, List[int], Tuple[int, ...]] = None
) -> Tensor:
    """
    :param tens: the tensor to calculate the sparsity for
    :param dim: the dimension(s) to split the calculations over;
        ex, can split over batch, channels, or combos
    :return: the sparsity of the input tens, ie the fraction of numbers that are zero
    """
    if dim is None:
        zeros = (tens.cpu() == 0).sum()
        total = tens.numel()

        return zeros.float() / float(total)

    if isinstance(dim, int):
        dim = [dim]

    if max(dim) >= len(tens.shape):
        raise ValueError(
            "Unsupported dim given of {} in {} for tensor shape {}".format(
                max(dim), dim, tens.shape
            )
        )

    sum_dims = [ind for ind in range(len(tens.shape)) if ind not in dim]
    zeros = (tens == 0).sum(dim=sum_dims) if sum_dims else tens == 0
    total = numpy.prod(
        [tens.shape[ind] for ind in range(len(tens.shape)) if ind not in dim]
    )

    permute_order = sorted(
        ((d, len(dim) - i - 1) for i, d in enumerate(dim)), reverse=True
    )
    permute = [d[1] for d in permute_order]

    if permute != [i for i in range(len(permute))]:
        # need to permute to get desired dimensions at the front
        zeros = zeros.permute(*permute).contiguous()

    return zeros.float() / float(total)


def tensor_density(tens: Tensor, dim: Union[None, int, Iterable[int]] = None) -> Tensor:
    """
    :param tens: the tensor to calculate the density for
    :param dim: the dimension(s) to split the calculations over; ex, can split over
        batch, channels, or combos
    :return: the density of the input tens, ie the fraction of numbers that are non zero
    """
    density = (tensor_sparsity(tens, dim) - 1.0) * -1.0

    return density


def tensor_sample(
    tens: Tensor,
    sample_size: int,
    dim: Union[None, int, List[int], Tuple[int, ...]] = None,
) -> Tensor:
    """
    :param tens: the tensor to grab samples from
    :param sample_size: the number of samples to grab overall if dim is not supplied
        or per each dim if it is
    :param dim: the dimension(s) to split the samples over;
        ex, can split over batch, channels, or combos
    :return: the sampled tensor
    """
    if sample_size < 1:
        raise ValueError("improper sample size given of {}".format(sample_size))

    if dim is None:
        indices = tens.new_zeros((sample_size,)).long().random_(0, tens.numel())
        samples = tens.view(-1)[indices]

        return samples

    if isinstance(dim, int):
        dim = [dim]

    if max(dim) >= len(tens.shape):
        raise ValueError(
            "Unsupported dim given of {} in {} for tensor shape {}".format(
                max(dim), dim, tens.shape
            )
        )

    if dim != [ind for ind in range(len(dim))]:
        # put the desired dimension(s) at the front to sample from
        tens = tens.permute(
            *dim, *[ind for ind in range(len(tens.shape)) if ind not in dim]
        )
        dim = [ind for ind in range(len(dim))]

    if not tens.is_contiguous():
        tens = tens.contiguous()

    num_indices = int(numpy.prod([tens.shape[ind] for ind in range(len(dim))]))
    elem_per_ind = int(
        numpy.prod([tens.shape[ind] for ind in range(len(dim), len(tens.shape))])
    )
    # create a new tensor with offsets set for each of our elements that we are indexing
    indices = tens.new_tensor(
        [ind * elem_per_ind for ind in range(num_indices)], dtype=torch.long
    ).unsqueeze(1)
    # now broadcast it across to the total number of elements we should end with
    indices = indices * tens.new_ones((num_indices, sample_size), dtype=torch.long)
    # finally add in a random number within the available range per index
    indices += tens.new_zeros((num_indices, sample_size), dtype=torch.long).random_(
        0, elem_per_ind
    )
    # get our samples
    samples = tens.view(-1)[indices.view(-1)]
    # reshape for the proper dimension
    samples = samples.view(*(tens.shape[ind] for ind in dim), sample_size)

    return samples


def tensor_list_sparsity(tensors: List[Tensor]) -> float:
    """
    :param tensors: the list of tensors to calculate the sparsity for
    :return: the total sparsity of all tensors in the list
    """
    zeros = 0
    numel = 0
    for tensor in tensors:
        zeros += (tensor == 0).sum().item()
        numel += tensor.numel()
    return float(zeros) / float(numel)


def mask_difference(old_mask: Tensor, new_mask: Tensor) -> Tensor:
    """
    :param old_mask: the old mask to compare against for calculating the difference
    :param new_mask: the new mask to compare with for calculating the difference
    :return: a tensor representing the change from the old_mask to the new_mask
             specifically values returned as 1.0 are newly unmasked (0.0 => 1.0)
             values returned as -1.0 are newly masked (1.0 => 0.0)
             values returned as 0.0 had no change in (0.0 => 0.0 or 1.0 => 1.0)
    """
    newly_masked = ((old_mask != new_mask) & (new_mask == 0.0)).type(old_mask.type())
    newly_unmasked = ((old_mask != new_mask) & (new_mask == 1.0)).type(old_mask.type())

    return -1.0 * newly_masked + newly_unmasked


##############################
#
# pytorch module helper functions
#
##############################


def get_layer(name: str, module: Module) -> Module:
    """
    :param name: the name of the layer to grab from the module
    :param module: the module containing the layer to grab
    :return: the module representing the layer in the module
    """
    if not name:
        return module

    layers = name.split(".")
    layer = module

    for name in layers:
        layer = layer.__getattr__(name)

    return layer


def get_terminal_layers(module: Module) -> Dict[str, Module]:
    """
    :param module: the module to grab all terminal layers for
    :return: a list of all of the terminal layers in a model
        (ie not containers; so convs, linears, activations, etc)
    """
    terminal = {}

    for mod_name, mod in module.named_modules():
        # check if it is a root node (only has itself in named_modules)
        child_count = 0
        for _, __ in mod.named_modules():
            child_count += 1

        if child_count != 1:
            continue

        terminal[mod_name] = mod

    return terminal


def get_conv_layers(module: Module) -> Dict[str, Module]:
    """
    :param module: the module to grab all conv layers for
    :return: a list of all the conv layers in the module
    """
    return {
        name: mod
        for name, mod in module.named_modules()
        if (isinstance(mod, _ConvNd) or (GPTConv1D and isinstance(mod, GPTConv1D)))
    }


def get_linear_layers(module: Module) -> Dict[str, Module]:
    """
    :param module: the module to grab all linear layers for
    :return: a list of all linear layers in the module
    """
    return {
        name: mod for name, mod in module.named_modules() if isinstance(mod, Linear)
    }


def get_prunable_layers(module: Module) -> List[Tuple[str, Module]]:
    """
    :param module: the module to get the prunable layers from
    :return: a list containing the names and modules of the prunable layers
        (Linear, ConvNd)
    """
    return [
        (name, mod)
        for (name, mod) in module.named_modules()
        if (
            isinstance(mod, Linear)
            or isinstance(mod, Embedding)
            or isinstance(mod, _ConvNd)
            or (QATLinear and isinstance(mod, QATLinear))
            or (QATConv2d and isinstance(mod, QATConv2d))
            or (QATConv3d and isinstance(mod, QATConv3d))
            or (GPTConv1D and isinstance(mod, GPTConv1D))
        )
    ]


def get_quantizable_layers(module: Module) -> List[Tuple[str, Module]]:
    """
    :param module: the module to get the quantizable layers from
    :return: a list containing the names and modules of the quantizable layers
        (Embedding, Linear, Conv2d, Conv3d)
    """
    if QATLinear is None:
        raise ImportError(
            "PyTorch version is not setup for Quantization. "
            "Please install a QAT compatible version of PyTorch"
        )

    return [
        (name, mod)
        for (name, mod) in module.named_modules()
        if (
            isinstance(mod, Linear)
            or isinstance(mod, Embedding)
            or isinstance(mod, Conv2d)
            or (QATConv3d and isinstance(mod, Conv3d))
        )
    ]


def get_quantized_layers(module: Module) -> List[Tuple[str, Module]]:
    """
    :param module: the module to get the quantized layers from
    :return: a list containing the names and modules of the quantized layers
        (Embedding, Linear, Conv2d, Conv3d)
    """

    quantized_layers = []
    for name, mod in module.named_modules():
        if hasattr(mod, "quantization_scheme"):
            weight_scheme = getattr(mod.quantization_scheme, "weights", None)
            if weight_scheme is not None and hasattr(mod, "weight"):
                quantized_layers.append((name, mod))

    return quantized_layers


def get_layer_param(param: str, layer: str, module: Module) -> Parameter:
    """
    :param param: the name of the param to grab from the layer
    :param layer: the name of the layer to grab from the module
    :param module: the module containing the layer and the param
    :return: the param taken from the given layer in the module
    """
    layer = get_layer(layer, module)  # type: Module
    param = layer.__getattr__(param)  # type: Parameter

    return param


def get_named_layers_and_params_by_regex(
    module: Module,
    param_names: List[str],
    params_strict: bool = False,
) -> List[NamedLayerParam]:
    """
    :param module: the module to get the matching layers and params from
    :param param_names: a list of names or regex patterns to match with full parameter
        paths. Regex patterns must be specified with the prefix 're:'
    :param params_strict: if True, this function will raise an exception if there a
        parameter is not found to match every name or regex in param_names
    :return: a list of NamedLayerParam tuples whose full parameter names in the given
        module match one of the given regex patterns or parameter names
    """
    named_layers_and_params = []
    found_param_names = []
    for layer_name, layer in module.named_modules():
        for param_name, param in layer.named_parameters():
            if "." in param_name:  # skip parameters of nested layers
                continue
            full_param_name = "{}.{}".format(layer_name, param_name)
            if any_str_or_regex_matches_param_name(full_param_name, param_names):
                named_layers_and_params.append(
                    NamedLayerParam(layer_name, layer, param_name, param)
                )
                found_param_names.append(full_param_name)
            elif layer_name.endswith(".module"):
                # unwrap layers wrapped with a QuantWrapper and check if they match
                parent_layer_name = ".".join(layer_name.split(".")[:-1])
                parent_layer = get_layer(parent_layer_name, module)
                skip_wrapper_name = "{}.{}".format(parent_layer_name, param_name)
                if (
                    QuantWrapper is not None
                    and isinstance(parent_layer, QuantWrapper)
                    and any_str_or_regex_matches_param_name(
                        skip_wrapper_name, param_names
                    )
                ):
                    named_layers_and_params.append(
                        NamedLayerParam(layer_name, layer, param_name, param)
                    )
                    found_param_names.append(skip_wrapper_name)
    if params_strict:
        validate_all_params_found(param_names, found_param_names)

    return named_layers_and_params


def any_str_or_regex_matches_param_name(
    param_name: str,
    name_or_regex_patterns: List[str],
) -> bool:
    """
    :param param_name: The name of a parameter
    :param name_or_regex_patterns: List of full param names to match to the input or
        regex patterns to match with that should be prefixed with 're:'
    :return: True if any given str or regex pattern matches the given name
    """
    for name_or_regex in name_or_regex_patterns:
        if name_or_regex[:3] == "re:":
            pattern = name_or_regex[3:]
            if re.match(pattern, param_name):
                return True
        else:
            if param_name == name_or_regex:
                return True
    return False


def validate_all_params_found(
    name_or_regex_patterns: List[str],
    found_param_names: List[str],
):
    """
    :param name_or_regex_patterns: List of full param names or regex patterns of them
        to check for matches in found_param_names names
    :param found_param_names: List of NamedLayerParam objects to check for matches
    :raise RuntimeError: If there is a name or regex pattern that does not have a
        match in found_param_names
    """
    for name_or_regex in name_or_regex_patterns:
        if "re:" != name_or_regex[:3] and name_or_regex in found_param_names:
            continue  # name found in list of full parameter names
        if "re:" == name_or_regex[:3] and any(
            re.match(name_or_regex[3:], name) for name in found_param_names
        ):
            continue  # regex pattern matches at least one full parameter name

        raise RuntimeError(
            "All supplied parameter names or regex patterns not found."
            "No match for {} in found parameters {}. \nSupplied {}".format(
                name_or_regex, found_param_names, name_or_regex_patterns
            )
        )


def set_deterministic_seeds(seed: int = 0):
    """
    Manually seeds the numpy, random, and torch packages.
    Also sets torch.backends.cudnn.deterministic to True
    :param seed: the manual seed to use. Default is 0
    """
    numpy.random.seed(seed)
    random.seed(seed)
    torch.manual_seed(seed)
    torch.backends.cudnn.deterministic = True


@contextmanager
def torch_distributed_zero_first(local_rank: Optional[int]):
    """
    Decorator to make all processes in distributed training wait for each
    local 0 ranked process to do something.
    :param local_rank: the local rank of this process
    """
    if local_rank is not None and local_rank not in [-1, 0]:
        torch.distributed.barrier()
    yield
    if local_rank == 0:
        torch.distributed.barrier()


def thin_model_from_checkpoint(model: Module, state_dict: Dict[str, Any]):
    """
    Updates any Linear/Conv/BN layers in the given model to match their
    respective shapes in the given state dict. Purpose of compatibility
    when loading weight for a model from a checkpoint of the same architecture
    but with potentially structured thinning applied. Note that this function
    has no guarantees on accuracy, will only resize model parameters for
    loading compatibility. All adjustments done in place

    :param model: model to potentially adjust parameter shapes of
    :param state_dict: state dict to infer parameter shapes from
    """
    first_thinned = True
    for param_name, checkpoint_tens in state_dict.items():
        if not param_name.endswith(".weight"):
            continue  # only deal with weight params of modules
        layer_name = param_name[:-7]
        layer = get_layer(layer_name, model)

        if not hasattr(layer, "weight") or (
            layer.weight.shape == checkpoint_tens.shape
        ):
            continue  # skip if there is no update to shape

        # quick check that target layer is some flavor of FC/Conv/BN
        layer_type = layer.__class__.__name__
        if not (
            "Linear" not in layer_type
            or "Conv" not in layer_type
            or ("BatchNorm" not in layer_type)
        ):
            continue

        orig_shape = layer.weight.shape
        target_shape = checkpoint_tens.shape

        # update weight param + grad
        if len(target_shape) > 1:
            layer.weight.data = layer.weight.data[
                : target_shape[0], : target_shape[1], ...
            ]
            if layer.weight.grad is not None:
                layer.weight.grad = layer.weight.grad[
                    : target_shape[0], : target_shape[1], ...
                ]
        else:
            layer.weight.data = layer.weight.data[: target_shape[0]]
            if layer.weight.grad is not None:
                layer.weight.grad = layer.weight.grad[: target_shape[0]]

        # update bias param + grad
        if hasattr(layer, "bias") and layer.bias is not None:
            # target output channels should be the first dim of target shape
            layer.bias.data = layer.bias.data[: target_shape[0]]
            if layer.bias.grad is not None:
                layer.bias.grad = layer.bias.grad[: target_shape[0]]

        # update layer attributes
        if "BatchNorm" in layer_type:
            if hasattr(layer, "num_features"):
                layer.num_features = layer.weight.size(0)
            # BN running mean and var are not stored as Parameters
            if hasattr(layer, "running_mean"):
                layer.running_mean = torch.zeros_like(layer.running_mean)[
                    : target_shape[0]
                ]
            if hasattr(layer, "running_var"):
                layer.running_var = torch.zeros_like(layer.running_var)[
                    : target_shape[0]
                ]

        if "Linear" in layer_type:
            if hasattr(layer, "out_features"):
                layer.out_features = layer.weight.shape[0]
            if hasattr(layer, "in_features"):
                layer.in_features = layer.weight.shape[1]

        if "Conv" in layer_type:
            if hasattr(layer, "out_channels"):
                layer.out_channels = layer.weight.shape[0]
            if hasattr(layer, "in_channels"):
                layer.in_channels = layer.weight.shape[1]
            if hasattr(layer, "groups") and layer.groups > 1:
                layer.groups = layer.weight.shape[0] // layer.weight.shape[1]

        if first_thinned:
            logger.info(
                "Thinning module layers for compatibility with given state dict:"
            )
            first_thinned = False
        logger.info(
            f"Thinned layer {layer_name} from shape {orig_shape} to "
            f"{layer.weight.shape}"
        )


##############################
#
# misc pytorch helper functions
#
##############################


MEMORY_BOUNDED = "MEMORY_BOUNDED"


def memory_aware_threshold(tensor: torch.Tensor, idx: int) -> Tensor:
    """
    Finds a threshold at the lookup idx in the most efficient way with available
    resources. Will be phased out when GPU-memory overhead of torch.sort reduces,
    or when torch.kthvalue becomes faster than torch.sort.

    :param tensor: A tensor to find a k-th smallest value in, where k=idx+1
    :param idx: A lookup index
    :return: k-th smallest value from the given tensor, where k=idx+1
    """
    try:
        if (
            MEMORY_BOUNDED in os.environ
            and os.environ[MEMORY_BOUNDED].lower() == "true"
        ):
            return torch.kthvalue(tensor.reshape(-1), idx + 1)[0]
        else:
            return torch.sort(tensor.reshape(-1))[0][idx]
    except RuntimeError:
        logger.warning(
            "Finding threshold from sparsity failed due to lack of memory, "
            "will attempt to recover. Consider setting env variable "
            f"{MEMORY_BOUNDED}=True in future runs."
        )
        torch.cuda.empty_cache()
        os.environ[MEMORY_BOUNDED] = "True"
        return torch.kthvalue(tensor.view(-1), idx + 1)[0]


def detach(x: Union[torch.Tensor, List, Tuple]):
    if isinstance(x, torch.Tensor):
        return x.detach()
    elif isinstance(x, List):
        return [detach(e) for e in x]
    elif isinstance(x, Tuple):
        return tuple([detach(e) for e in x])
    else:
        raise ValueError("Unexpected type to detach")


def adjust_quantization_for_onnx_export(module: torch.nn.Module) -> torch.nn.Module:
    # supported pytorch ranges are int8 or uint8
    allowed_ranges = [(0, 127), (0, 255), (-128, 127)]
    fake_quant_modules = [
        m for m in module.modules() if m.__class__.__name__ == "FakeQuantize"
    ]

    if _PARSED_TORCH_VERSION >= version.parse("1.12"):
        for quant in fake_quant_modules:
            # original ranges preserved in quant.quant_min and quant.quant_max
            quant_range = (
                quant.activation_post_process.quant_min,
                quant.activation_post_process.quant_max,
            )
            if quant_range not in allowed_ranges:
                if quant_range[0] < 0:  # convert signed range to int8
                    quant.activation_post_process.quant_min = -128
                    quant.activation_post_process.quant_max = 127
                else:  # convert unsigned range to uint8
                    quant.activation_post_process.quant_min = 0
                    quant.activation_post_process.quant_max = 255
            # don't update observer since ranges are artificially modified
            quant.observer_enabled[0] = 0

    else:  # backwards compatibility for torch <= 1.11
        for quant in fake_quant_modules:
            quant_range = (quant.quant_min, quant.quant_max)
            if quant_range not in allowed_ranges:
                if quant_range[0] < 0:  # convert signed range to int8
                    quant.quant_min = -128
                    quant.quant_max = 127
                else:  # convert unsigned range to uint8
                    quant.quant_min = 0
                    quant.quant_max = 255
            # don't update observer since ranges are artificially modified
            quant.observer_enabled[0] = 0


def get_dependency_order(
    layer: Module, subset: Dict, an_input: Tensor, **kwargs
) -> List[str]:
    """
    Get a list of a subset of modules in layer ordered by execution order, which honors
    the dependencies in the graph

    :param layer: pytorch module to calculate dependencies for
    :param subset: subset of modules in the layer to include in the ordering
    :param an_input: example input to pass through the layer forward pass, used to
        determine execution order

    :return: list of module names in execution order
    """
    order = []

    def exe_input(name):
        def _exe_input(_, inp, out):
            if name in subset:
                order.append(name)

        return _exe_input

    # register a hook for each module of interest, will be triggered in exeuction order
    handles = [subset[name].register_forward_hook(exe_input(name)) for name in subset]
    layer(an_input, **kwargs)
    for h in handles:
        h.remove()
    return order


def swap_modules(
    module: torch.nn.Module, submodule_name: str, submodule_to_replace: torch.nn.Module
) -> torch.nn.Module:
    """
    Iteratively unfold the submodules of the module according to the submodule_name
    to eventually replace the leaf submodule (accessed from the module through the
    submodule_name) with the submodule_to_replace.

    E.g
    ```
    swap_modules(module=Model,
                 module_name="layers.0.sublayer",
                 module_to_replace=ReplaceModule
                 )
    ```
    this will iteratively traverse through the submodules
    'layers' -> '0' -> to eventually replace 'sublayer' with ReplaceModule

    :param module: the module to replace with the module_to_replace
    :param submodule_name: the name of the module to replace
    :param submodule_to_replace: the module to replace the module with
    :return: the replaced module
    """
    parent = module
    sections = submodule_name.split(".")

    for sec in sections[:-1]:
        parent = parent.__getattr__(sec)

    cur = parent.__getattr__(sections[-1])
    parent.__setattr__(sections[-1], submodule_to_replace)

    return cur



from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

import numpy as np
import torch
from loguru import logger
from torch.nn import Module
from tqdm import tqdm

from llmcompressor.core.state import State
from llmcompressor.modifiers import Modifier
from llmcompressor.modifiers.pruning.wanda.utils.wanda_wrapper import WandaWrapper
from llmcompressor.modifiers.utils.layer_compressor import LayerCompressor
from llmcompressor.modifiers.utils.pytorch_helpers import run_calibration_forward
from llmcompressor.utils.pytorch.module import (
    get_layers,
    get_no_split_params,
    get_prunable_layers,
)

__all__ = ["WandaPruningModifier"]


class WandaPruningModifier(Modifier):
    """
    Modifier for applying the one-shot WANDA algorithm to a model
    from the paper: https://arxiv.org/abs/2306.11695

    Lifecycle:
        - on_initialize
            - initialize_compression()
                - compressible_layers()
                - LayerCompressor.pre_compress()
            - apply_compression()
                - run_calibration_forward()
                - LayerCompressor.compress()
                - LayerCompressor.post_compress()
                - LayerCompressor.revert_layer_wrappers()
        - on_finalize

    :param sparsity: Sparsity to compress model to
    :param mask_structure: String to define the structure of the mask to apply.
        Must be of the form N:M where N, M are integers that define a custom block
        shape. Defaults to 0:0 which represents an unstructured mask.
    :param sequential_update: Whether or not to update weights sequentially by layer,
        True saves on GPU memory
    :param targets: list of layer names to compress during OBCQ, or '__ALL__'
        to compress every layer in the model
    """

    sparsity: Union[float, List[float]] = 0.0
    sparsity_profile: Optional[str] = None
    owl_m: Optional[int] = None
    owl_lmbda: Optional[float] = None
    mask_structure: str = "0:0"
    sequential_update: Optional[bool] = False
    targets: Union[str, List[str], None] = None
    model: Optional[Any] = None
    layer_compressors_: List = None

    compressible_layers_: Optional[List] = None
    prunen_: Optional[int] = None
    prunem_: Optional[int] = None

    def on_initialize_structure(self, state: State, **kwargs):
        """
        This modifier does not alter the model structure.
        This method is a no-op.

        :param state: Unused, kept to conform to the parent method signature
        :param kwargs: Unused, kept to conform to the parent method signature
        """

    def on_initialize(self, state: State, **kwargs) -> bool:
        """
        Initialize and run the WANDA algorithm on the current state

        :param state: session state storing input model and calibration data
        :param kwargs: Unused, kept to conform to the parent method signature
        """
        modifiable_model = state.model
        calibration_dataloader = state.data.calib

        if self.targets is None:
            # if no targets are provided, default to the modules that shouldn't be
            # split by FSDP. For Transformers models this is equivalent to the
            # decoder layers (ie LlamaDecoderLayer)
            self.targets = get_no_split_params(modifiable_model)

        self.initialize_compression(modifiable_model, calibration_dataloader)
        self.apply_compression(calibration_dataloader)

        return True

    def on_finalize(self, state: State, **kwargs):
        """
        Nothing to clean up for this module

        :param state: Unused, kept to conform to the parent method signature
        :param kwargs: Unused, kept to conform to the parent method signature
        """

        return True

    def compressible_layers(self) -> Dict:
        """
        Retrieves the modules corresponding to a list of
        compressible layer names

        :precondition: self.model is set and is a torch.nn.Module
        :return: dictionary of modules to compress
        """
        if not isinstance(self.model, Module):
            raise ValueError(
                "`self.model` must be a torch.nn.Module to use "
                f"the {self.__class__.__qualname__} modifier but got "
                f"{type(self.model)} instead"
            )

        return get_layers(self.targets, self.model)

    def initialize_compression(
        self,
        model: Module,
        dataloader: Optional[Iterable[Tuple[List, Dict[str, Any]]]] = None,
    ):
        """
        Setup for WANDA, initializes the model, device,
        and other parameters, also initilializes the
        compressible layers of model, and sets the device

        :param model: model to initialize for compression
        """
        self.model = model
        self.compressible_layers_ = self.compressible_layers()
        self.layer_compressors_ = []
        self._infer_mask_block_size()

        if self.sparsity_profile is not None and self.sparsity_profile.lower() == "owl":
            logger.info(
                "Inferring layer-wise sparsities from "
                f"{len(dataloader) if dataloader else 0} calibration samples..."
            )
            self.sparsity = self._infer_layer_sparsity(dataloader)
        self._validate_layerwise_sparsity()

        for idx, (name, layer) in enumerate(self.compressible_layers_.items()):
            logger.info(f"Preparing {name} for compression")
            if isinstance(self.sparsity, Dict):
                layer_sparsity = self.sparsity[name]
            elif isinstance(self.sparsity, List):
                layer_sparsity = self.sparsity[idx]
            else:  # float
                layer_sparsity = self.sparsity
            args = self._pruning_arguments(layer_sparsity)
            comp_cls = self._compression_class()
            compressor = LayerCompressor(comp_cls, self.model, layer, idx, name, args)
            if not self.sequential_update:
                # add all batch processing hooks before the forward pass
                compressor.pre_compress()
            self.layer_compressors_.append(compressor)

    @torch.no_grad()
    def apply_compression(
        self, dataloader: Optional[Iterable[Tuple[List, Dict[str, Any]]]] = None
    ) -> Dict:
        """
        Run Wanda on the loaded model, using dataloader as calibration data

        :param dataloader: calibration data for WANDA
        """
        class_name = self.__class__.__name__.replace("PyTorch", "")
        logger.info(
            f"Running {class_name} calibration with " f"{len(dataloader)} samples..."
        )
        if not self.sequential_update:
            # in non-sequential mode we run one forward batch for all modules
            run_calibration_forward(self.model, dataloader, mask_padding=True)

        num_layers = len(self.compressible_layers_)
        for idx, layer_compressor in enumerate(self.layer_compressors_):
            layer_sparsity = layer_compressor.args["sparsity"]
            logger.info(
                f"\n===== Compressing layer {idx+1}/{num_layers} "
                f"to sparsity {layer_sparsity} ====="
            )

            # Prune/quantize using the layer compressor
            if self.sequential_update:
                # in sequential mode we run one forward pass for each module we
                # want to compress, this will be really slow but allows compression in
                # earlier layers to affect later layers
                layer_compressor.pre_compress()
                logger.info(f"Calibrating {layer_compressor.name}...")
                run_calibration_forward(self.model, dataloader, mask_padding=True)
            layer_compressor.compress()
            layer_compressor.post_compress()
            layer_compressor.revert_layer_wrappers()
            torch.cuda.empty_cache()

    def _validate_layerwise_sparsity(self):
        if isinstance(self.sparsity, float):
            # single sparsity will be applied to all layers
            return

        target_layers = list(self.compressible_layers_.keys())

        if len(target_layers) != len(self.sparsity):
            raise ValueError(
                "Number of layer targets must match the number of "
                f"sparsities. Got {len(target_layers)} layers and "
                f"{len(self.sparsity)} sparsities"
            )

    def _pruning_arguments(self, sparsity) -> Dict[str, Any]:
        """
        Gather the parameters needed for root module compression in a dict

        :param sparsity: target sparsity
        :return: dict of params for pruning
        """
        return {
            "sparsity": sparsity,
            "prunen": self.prunen_,
            "prunem": self.prunem_,
        }

    def _compression_class(self):
        """
        :return: wrapper class used for root modules of this compression class
        """
        return WandaWrapper

    def _infer_mask_block_size(self):
        """
        Infer the mask block size from the mask structure.
        Parses mask_structure of the form N:M where N, M are integers that
        define a custom block shape; and sets prunen_ and prunem_ accordingly.

        :post-condition: prunen_ and prunem_ are set
        """
        if self.mask_structure is None:
            raise ValueError("mask_structure must be defined")

        self.prunen_, self.prunem_ = list(map(int, self.mask_structure.split(":")))

    def _infer_layer_sparsity(self, calibration_dataloader):
        acts = _get_activations(self.model, calibration_dataloader)
        wanda = {}
        for name, layer in self.compressible_layers_.items():
            prunable_layers = get_prunable_layers(layer)
            z = [
                m.weight.abs() * acts[f"{name}.{n}"].unsqueeze(0)
                for n, m in prunable_layers.items()
            ]
            wanda[name] = torch.cat([item.flatten().cpu() for item in z])

        acts = None
        del acts
        torch.cuda.empty_cache()

        outlier_ratios = {}
        for group in wanda:
            threshold = torch.mean(wanda[group]) * self.owl_m
            outlier_ratios[group] = (
                100 * (wanda[group] > threshold).sum().item() / wanda[group].numel()
            )
        outlier_ratios_arr = np.array([outlier_ratios[k] for k in outlier_ratios])
        for k in outlier_ratios:
            outlier_ratios[k] = (outlier_ratios[k] - outlier_ratios_arr.min()) * (
                1
                / (outlier_ratios_arr.max() - outlier_ratios_arr.min())
                * self.owl_lmbda
                * 2
            )
        outlier_ratios_arr = np.array([outlier_ratios[k] for k in outlier_ratios])
        sparsities = {
            k: 1
            - (
                outlier_ratios[k]
                - np.mean(outlier_ratios_arr)
                + (1 - float(self.sparsity))
            )
            for k in outlier_ratios
        }
        logger.info(f"OWL sparsities for sp={self.sparsity} are:")
        for k in sparsities:
            logger.info(f"Sparsity for {k}: {sparsities[k]}")
        return sparsities


@torch.no_grad()
def _get_activations(model, data_loader, nsamples=128):
    import functools

    model.eval()
    acts = {}

    def save_acts(module, input, name):
        if isinstance(input, tuple):
            input = input[0]
        if name not in acts:
            acts[name] = 1.0 / nsamples * input.detach().pow(2).sum(dim=(0, 1)).sqrt()
        else:
            acts[name] += 1.0 / nsamples * input.detach().pow(2).sum(dim=(0, 1)).sqrt()

    hooks = []
    for name, mod in model.named_modules():
        if isinstance(mod, torch.nn.Linear) and "lm_head" not in name:
            hooks.append(
                mod.register_forward_pre_hook(functools.partial(save_acts, name=name))
            )
    device = next(model.parameters()).device
    for batch in tqdm(data_loader):
        batch = {k: v.to(device) for k, v in batch.items()}
        model(**batch)
        batch = None
    torch.cuda.empty_cache()

    for h in hooks:
        h.remove()

    return acts



from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

import torch
from compressed_tensors.quantization import (
    QuantizationScheme,
    disable_quantization,
    enable_quantization,
    freeze_module_quantization,
)
from loguru import logger
from pydantic import Field
from torch.nn import Module

from llmcompressor.core.state import State
from llmcompressor.modifiers import Modifier, ModifierFactory
from llmcompressor.modifiers.quantization.gptq.utils.gptq_wrapper import GPTQWrapper
from llmcompressor.modifiers.utils.layer_compressor import LayerCompressor
from llmcompressor.modifiers.utils.pytorch_helpers import run_calibration_forward
from llmcompressor.utils.fsdp.context import fix_fsdp_module_name
from llmcompressor.utils.pytorch.module import (
    get_layers,
    get_no_split_params,
    qat_active,
)

__all__ = ["GPTQModifier"]


class GPTQModifier(Modifier):
    """
    Modifier for applying the one-shot OBCQ algorithm to a model

    Lifecycle:
        - on_initialize
            - initialize_compression()
                - compressible_layers()
                - LayerCompressor.pre_compress()
            - apply_compression()
                - run_calibration_forward()
                - LayerCompressor.compress()
                - LayerCompressor.post_compress()
                - LayerCompressor.revert_layer_wrappers()
    | Sample yaml:
    | test_stage:
    |    obcq_modifiers:
    |      GPTQModifier:
    |          sequential_update: true
    |          dampening_frac: 0.001
    |          block_size: 128
    |          config_groups:
    |            group_0:
    |                targets:
    |                  - "Linear"
    |                input_activations: null
    |                output_activations: null
    |                weights:
    |                    num_bits: 8
    |                    type: "int"
    |                    symmetric: true
    |                    strategy: "tensor"
    |                    group_size: 128
    |                    actorder: False


    :param sequential_update: Whether or not to update weights sequentially by layer,
        True saves on GPU memory, default is True
    :param targets: list of layer names to compress during GPTQ, or '__ALL__'
        to compress every layer in the model
    :param block_size: Used to determine number of columns to compress in one pass
    :param quantize: Set to True to quantize using an existing quantization modifier,
        or pass in the configuration for a quantization modifier if one does not
        already exist in the recipe
    :param dampening_frac: Amount of dampening to apply to H, as a fraction of the
        diagonal norm
    :param config_groups: [Used, if a quantization modifier is not specified],
        dictionary specifying quantization schemes to apply to target
        modules. Modules not matching a scheme target will NOT be quantized.
    :param ignore: [Used, if a quantization modifier is not specified]
        optional list of module class names or submodule names to not
        quantize even if they match a target in config_groups. Defaults to empty list.
    :param disable_quantization_observer_epoch: [Used, if a quantization modifier is
        not specified] Epoch to disable updates to the module
        quantization observers. At this point, quantized weights and zero points will
        not be updated. Leave None to not disable observers during QAT. Default is None
    :param num_calibration_steps: Number of steps to run post training calibration for.
        When None, the entire calibration_dataloader is used
    :param scheme: [Used, if a quantization modifier is not specified], the quantization
        scheme to apply to the model, this is a dictionary that supports all keys from
        QuantizationScheme except targets, which will be set to the targets parameter
        set at the modifier level. Can also be set to a dictionary of the format
        `preset_scheme_name: targets` for example: `W8A8: ['Linear']` for weight 8 bit
        or a string of a preset scheme if targets is provided
        and activation 8 bit quantization on the Linear layers.
    """

    sequential_update: Optional[bool] = True
    targets: Union[str, List[str], None] = None
    sequential_targets: Union[str, List[str], None] = None
    block_size: int = 128
    quantize: Union[bool, Dict] = True
    dampening_frac: Optional[float] = 0.01
    config_groups: Optional[Dict[str, QuantizationScheme]] = None
    ignore: List[str] = Field(default_factory=list)
    disable_quantization_observer_epoch: Optional[float] = None
    num_calibration_steps: Optional[int] = None
    scheme: Optional[Union[str, Dict[str, Any]]] = None

    model: Optional[Any] = None
    layer_compressors_: Optional[List[Any]] = None
    compressible_layers_: Optional[List] = None
    quantization_modifier_: Any = None

    def on_initialize_structure(self, state: State, **kwargs):
        """
        Check the model's quantization state matches that expected by this modifier,
        adding a default quantization scheme if needed

        :param state: session state storing input model and calibration data
        """
        quantization_already_active = qat_active(state.model)
        if isinstance(self.quantize, bool):
            if not self.quantize and quantization_already_active:
                logger.warning(
                    "GPTQ quantization is set to False, but a "
                    "quantization modifier is already active on the model "
                    "resetting quantize to True"
                )
                self.quantize = True
            elif self.quantize and not quantization_already_active:
                logger.warning(
                    "GPTQ quantization is set to True without an "
                    "active quantization modifier."
                )
                self._build_quant_modifier()
            return  # use existing quantization modifier if there is one
        else:
            if not isinstance(self.quantize, Dict):
                raise ValueError(
                    "GPTQModifier.quantize accepts only a single "
                    "quantization modifier or a boolean. Found "
                    f"type {type(self.quantize)}"
                )
            if len(self.quantize) != 1:
                raise ValueError(
                    "GPTQModifier.quantize accepts only a single "
                    "quantization modifier or a boolean. Found "
                    f"{len(self.quantize)} modifiers"
                )
            if quantization_already_active:
                logger.warning(
                    "Attempting to initialize quantization for GPTQ "
                    "but a quantization modifier has already been applied. "
                    "The quantization configuration defined under the "
                    "GPTQ modifier will be ignored."
                )
                self.quantize = True
                return
            self._build_quant_modifier_from_dict(self.quantize)
            self.quantize = True

        if self.quantization_modifier_:
            self.quantization_modifier_.on_initialize_structure(state, **kwargs)

    def on_initialize(self, state: "State", **kwargs) -> bool:
        """
        Initialize and run the GPTQ algorithm on the current state

        :param state: session state storing input model and calibration data
        """
        if not self.initialized_structure_:
            self.on_initialize_structure(state, **kwargs)
        if self.quantization_modifier_:
            self.quantization_modifier_.initialize(state, **kwargs)
        if not self.quantize:
            raise ValueError("To use the GPTQModifier, quantization must be enabled.")

        modifiable_model = state.model
        calibration_dataloader = state.data.calib

        if self.sequential_targets is None:
            # if no targets are provided, default to the modules that shouldn't be
            # split by FSDP. For Transformers models this is equivalent to the
            # decoder layers (ie LlamaDecoderLayer)
            self.sequential_targets = get_no_split_params(modifiable_model)

        self.initialize_compression(modifiable_model, calibration_dataloader)
        self.apply_compression(calibration_dataloader)
        state.model.apply(freeze_module_quantization)

        return True

    def on_finalize(self, state: "State", **kwargs) -> bool:
        """
        disable the quantization observers used by the OBCQ algorithm

        :param state: session state storing input model and calibration data
        """
        if self.quantization_modifier_:
            self.quantization_modifier_.finalize(state, **kwargs)

        return True

    def compressible_layers(self) -> Dict:
        """
        Retrieves the modules corresponding to a list of
        compressible layer names

        :precondition: self.model is set and is a torch.nn.Module
        :return: dictionary of modules to compress
        """
        if not isinstance(self.model, Module):
            raise ValueError(
                "`self.model` must be a torch.nn.Module to use "
                f"the {self.__class__.__qualname__} modifier but got "
                f"{type(self.model)} instead"
            )

        return get_layers(self.sequential_targets, self.model)

    def initialize_compression(
        self,
        model: Module,
        dataloader: Optional[Iterable[Tuple[List, Dict[str, Any]]]] = None,
    ):
        """
        Setup for GPTQ, initializes the model
        and other parameters, also initilializes the
        compressible layers of model, and sets the device

        :param model: model to initialize for compression
        :param dataloader: calibration data for GPTQ
        """
        self.model = model
        self.compressible_layers_ = self.compressible_layers()
        self.layer_compressors_ = []

        for idx, (name, layer) in enumerate(self.compressible_layers_.items()):
            name = fix_fsdp_module_name(name)
            logger.info(f"Preparing {name} for compression")
            args = self._pruning_arguments()
            comp_cls = self._compression_class()
            compressor = LayerCompressor(comp_cls, self.model, layer, idx, name, args)
            if not self.sequential_update:
                # add all batch processing hooks before the forward pass
                compressor.pre_compress()
            self.layer_compressors_.append(compressor)

        if self.sequential_update:
            first_layer_compressor = self.layer_compressors_[0]
            first_layer_compressor.set_early_stop()

    @torch.no_grad()
    def apply_compression(
        self, dataloader: Optional[Iterable[Tuple[List, Dict[str, Any]]]] = None
    ) -> Dict:
        """
        Run GPTQ on the loaded model, using dataloader as calibration data

        :param dataloader: calibration data for GPTQ
        """
        class_name = self.__class__.__name__.replace("PyTorch", "")
        logger.info(
            f"Running {class_name} calibration with " f"{len(dataloader)} samples..."
        )

        # quantization scales and zp are already initialized but we do not
        # want to calibrate wrt to these
        self.model.apply(disable_quantization)

        forward_pass_use_cache = self.model.config.use_cache
        self.model.config.use_cache = False

        # in non-sequential mode we run calibration through the full model
        # in sequential mode we run calibration up to the first transformer target
        intermediates = run_calibration_forward(
            self.model, dataloader, mask_padding=True
        )
        self.layer_compressors_[0].clear_early_stop()

        num_layers = len(self.compressible_layers_)
        for idx, layer_compressor in enumerate(self.layer_compressors_):
            logger.info(f"\n===== Compressing layer {idx+1}/{num_layers} " " =====")

            # Prune/quantize using GPTQ
            if self.sequential_update:
                # in sequential mode we run the forward pass for each transformer layer
                # one at a time, caching the intermediate outputs between layers
                layer_compressor.pre_compress()
                logger.info(f"Calibrating {layer_compressor.name}...")
                intermediates = layer_compressor.calibrate_layer(intermediates)

            layer_compressor.compress()
            layer_compressor.post_compress()
            layer_compressor.revert_layer_wrappers()
            torch.cuda.empty_cache()

        self.model.config.use_cache = forward_pass_use_cache

        # re-enable quantization
        self.model.apply(enable_quantization)

    def _build_quant_modifier(self):
        """
        Build a quantization modifier based on the specified config_groups,
        ignore list, and num_calibration_steps.

        :postcondition: self.quantization_modifier_ is set to the built
            quantization modifier
        """

        quantization_args_names = [
            "config_groups",
            "targets",
            "scheme",
            "num_calibration_steps",
            "ignore",
            "disable_quantization_observer_epoch",
        ]

        quant_args = {
            key: getattr(self, key)
            for key in quantization_args_names
            if getattr(self, key, False)
        }

        logger.info(f"Building quantization modifier with args: {quant_args}")
        vllm_quant_config = {"QuantizationModifier": quant_args}
        self._build_quant_modifier_from_dict(vllm_quant_config)

    def _build_quant_modifier_from_dict(self, quant_config):
        modifier_type = list(quant_config.keys())[0]
        modifier_args = quant_config[modifier_type]
        self.quantization_modifier_ = ModifierFactory.create(
            modifier_type,
            allow_registered=True,
            allow_experimental=True,
            **modifier_args,
        )

    def _pruning_arguments(self):
        """
        Gather the parameters needed for root module compression in a dict

        :return: dict of params for pruning
        """
        return {
            "blocksize": self.block_size,
            "percdamp": self.dampening_frac,
        }

    def _compression_class(self):
        """
        :return: wrapper class used for root modules of this compression class
        """
        return GPTQWrapper



from llmcompressor.transformers.finetune.data import TextGenerationDataset


@TextGenerationDataset.register(name="wikitext")
class WikiTextDataset(TextGenerationDataset):
    """
    Child text generation class for the Open Platypus dataset

    :param data_args: configuration settings for dataset loading
    :param split: split from dataset to load, for instance `test` or `train[:5%]`
    :param tokenizer: tokenizer to use on dataset
    """

    def __init__(self, data_args, split, tokenizer):
        super().__init__(
            text_column="text", data_args=data_args, split=split, tokenizer=tokenizer
        )



from typing import Dict, List, Optional, Union

import psutil
import torch
from accelerate import infer_auto_device_map, init_empty_weights
from accelerate.accelerator import get_state_dict_offloaded_model
from torch.nn.modules import Linear
from tqdm import tqdm
from transformers import AutoModelForCausalLM

from llmcompressor.pytorch.utils import get_linear_layers
from llmcompressor.utils.pytorch import get_layers, get_no_split_params

__ALL__ = [
    "tensor_follows_mask_structure",
    "infer_sparsity_structure_from_stage_modifiers",
    "infer_sparsity_structure_from_model",
    "hessian_memory_requirements",
    "custom_offload_device_map",
    "calculate_offload_device_map",
]


def tensor_follows_mask_structure(tensor, mask: str = "2:4") -> bool:
    """
    :param tensor: tensor to check
    :param mask: mask structure to check for, in the format "n:m"
    :return: True if the tensor follows the mask structure, False otherwise.
        Note, some weights can incidentally be zero, so we check for
        atleast n zeros in each chunk of size m
    """

    n, m = tuple(map(int, mask.split(":")))
    # Reshape the tensor into chunks of size m
    tensor = tensor.view(-1, m)

    # Count the number of zeros in each chunk
    zero_counts = (tensor == 0).sum(dim=1)

    # Check if the number of zeros in each chunk atleast n
    # Greater than sign is needed as some weights can incidentally
    # be zero
    return torch.all(zero_counts >= n).item()


def infer_sparsity_structure_from_stage_modifiers(
    stage_modifiers: List["StageModifier"],  # noqa E501
) -> Optional[str]:
    """
    Determines the sparsity structure, if any exists, given the
    list of stage modifiers

    :param stage_modifiers: non-empty list of stage modifiers
    :return: sparsity structure as a string or None
    """
    for stage in stage_modifiers:
        if stage.applied:
            for modifier in stage.modifiers:
                if hasattr(modifier, "mask_structure"):
                    sparsity_structure = modifier.mask_structure
                    return sparsity_structure
    return None


def infer_sparsity_structure_from_model(model: torch.nn.Module) -> Optional[str]:
    """
    Determines the sparsity structure, if any exists, given the model

    :param model: model to check for sparsity structure
    :return: sparsity structure as a string or None
    """

    # check for the common sparsity structures
    structures = {"2:4"}
    for sparsity_structure in structures:
        linear_modules = get_linear_layers(model)
        offloaded_params = get_state_dict_offloaded_model(model)

        linear_modules_with_sparsity_structure = [
            tensor_follows_mask_structure(offloaded_params[f"{name}.weight"])
            for name in tqdm(
                linear_modules.keys(),
                desc="Checking whether model follows "
                f"{sparsity_structure} sparsity structure",
            )
        ]
        # if the majority of the linear modules follow the sparsity structure
        # we can assume that the model follows the sparsity structure
        # (taking into consideration the fact that some Linear layers like the
        # embedding layer might not be sparse)
        if (
            sum(linear_modules_with_sparsity_structure)
            > len(linear_modules_with_sparsity_structure) * 0.8
        ):
            return sparsity_structure

    return None


def hessian_memory_requirements(model: torch.nn.Module) -> int:
    """
    Determines the number of bytes needed to store Hessian data for a single
    transformer layer in model. This is used for reserving memory for GPTQ
    quantization

    :param model: model to calculate requirements for
    :return: number of bytes required to reserve for GPTQ on a single layer
    """
    transformer_layers = get_layers(get_no_split_params(model), model)
    total_hessian_elems = {}
    max_column_size = {}
    for no_split_name, no_split_layer in transformer_layers.items():
        total_hessian_elems[no_split_name] = 0
        max_column_size[no_split_name] = 0
        for name, module in no_split_layer.named_modules():
            if isinstance(module, Linear):
                for param in module.parameters():
                    column_size = param.shape[1]
                    total_hessian_elems[no_split_name] += column_size * column_size
                    if column_size > max_column_size[no_split_name]:
                        # max extra memory for inverse calculation
                        max_column_size[no_split_name] = column_size

    max_total_hessian_elems = max(total_hessian_elems.values())
    overall_max_column_size = max(max_column_size.values())
    bytes_per_weight = 32 // 8  # hessians are float32
    inverse_reserved = overall_max_column_size * overall_max_column_size
    return (max_total_hessian_elems + inverse_reserved) * bytes_per_weight


def quantization_memory_requirement(model: torch.nn.Module) -> int:
    """
    Determines the max number of bytes needed to store quantization scale and zp data

    :param model: model to calculate requirements for
    :return: number of bytes required to reserve for quantization
    """

    total_elements = 0
    for _, module in model.named_modules():
        if isinstance(module, Linear):
            for param in module.parameters():
                # assume the max of group 128 and static scale/zp
                # TODO: base this on the recipe instead instead of assuming max

                # potentially just bias term
                max_quant_shape = param.shape[0] // 128

                if len(param.size()) > 1:  # weights
                    max_quant_shape *= param.shape[1]

                total_elements += max_quant_shape * 4

    bytes_ratio = 32 // 16  # assuming float16
    return total_elements * bytes_ratio


def custom_offload_device_map(
    model_stub: str,
    max_memory_per_gpu: Union[str, int],
    num_gpus: int = 1,
    **model_kwargs,
) -> Dict[Union[int, str], Union[int, str]]:
    """
    Calculates the optimal gpu mappings for model_stub stored as torch_dtype, where
    each GPU is restricted to allocating a specific amount of memory.

    :param model_stub: local path or HF stub to calculate mapping for
    :param max_memory_per_gpu: Max memory to allocate on each GPU, as either a string
        such as "10GB" or an integer number of bytes
    :param num_gpus: number of gpus to utilize
    :param model_kwargs: keyword arguments to pass to model initializer
    :return: memory mapping for layers of model_stub to be passed to from_pretrained()
    """
    max_cpu_memory = psutil.virtual_memory().available
    memory_limits = {device: max_memory_per_gpu for device in range(num_gpus)}
    memory_limits["cpu"] = max_cpu_memory

    device_map = {}
    with init_empty_weights():
        dummy_model = AutoModelForCausalLM.from_pretrained(model_stub, **model_kwargs)
        device_map = infer_auto_device_map(
            dummy_model,
            max_memory=memory_limits,
            no_split_module_classes=dummy_model._no_split_modules,
        )
        del dummy_model

    return device_map


def calculate_offload_device_map(
    model_stub: str,
    reserve_for_hessians=False,
    num_gpus: int = 1,
    torch_dtype: torch.dtype = torch.float16,
    **model_kwargs,
) -> Dict[Union[int, str], Union[int, str]]:
    """
    Calculates the optimal gpu mappings for model_stub stored as torch_dtype. Takes
    into account extra memory required for quantization and (optionally) GPTQ hessians

    :param model_stub: local path or HF stub to calculate mapping for
    :param reserve_for_hessians: whether to reserve memory for GPTQ
    :param num_gpus: number of gpus to utilize
    :param model_kwargs: keyword arguments to pass to model initializer
    :return: memory mapping for layers of model_stub to be passed to from_pretrained()
    """
    max_cpu_memory = psutil.virtual_memory().available
    max_gpu_memory = torch.cuda.mem_get_info(0)[0]
    available_gpus = torch.cuda.device_count()
    if available_gpus < num_gpus:
        raise ValueError(
            f"Requested {num_gpus} GPUs but only {available_gpus} are available."
        )
    max_gpu_memory = [max_gpu_memory] * num_gpus

    device_map = {}
    with init_empty_weights():
        dummy_model = AutoModelForCausalLM.from_pretrained(
            model_stub, torch_dtype=torch_dtype, **model_kwargs
        )

        reserved_memory = 0
        if reserve_for_hessians:
            reserved_memory = hessian_memory_requirements(dummy_model)
        reserved_memory += quantization_memory_requirement(dummy_model)

        memory_limits = {
            idx: (max_memory - reserved_memory)
            for idx, max_memory in enumerate(max_gpu_memory)
        }
        memory_limits["cpu"] = max_cpu_memory

        device_map = infer_auto_device_map(
            dummy_model,
            max_memory=memory_limits,
            no_split_module_classes=dummy_model._no_split_modules,
        )
        del dummy_model

    return device_map



import time
from typing import Tuple

from compressed_tensors.quantization import (
    ActivationOrdering,
    QuantizationArgs,
    QuantizationStrategy,
)
from compressed_tensors.quantization.lifecycle.forward import fake_quantize

from llmcompressor.modifiers.utils import SPARSITY_THRESHOLD
from llmcompressor.modifiers.utils.compression_wrapper import ModuleCompressionWrapper
from llmcompressor.pytorch.utils.helpers import tensor_sparsity
from llmcompressor.utils import getattr_chain
from llmcompressor.utils.metric_logging import (
    get_GPU_memory_usage,
    get_layer_size_bytes,
)

try:
    import transformers
except ImportError as err:
    transformers = None
    transformers_err = err

import math
from copy import copy

import torch
import torch.nn as nn
from compressed_tensors.utils import (
    get_offloaded_device,
    is_module_offloaded,
    update_parameter_data,
    update_prefix_dict,
)
from loguru import logger

__all__ = ["GPTQWrapper"]


class GPTQWrapper(ModuleCompressionWrapper):
    """
    Runs GPTQ on a single module that contains no sub-modules

    Lifecycle:
        - add_batch
        - compress
        - free

    :param name: name of module to run compression on
    :param layer: module to run compression on
    """

    def __init__(self, name, layer):
        super().__init__(name=name, layer=layer)

        # for Hessian calculation
        self.register_buffer(
            "H", torch.zeros((self.columns, self.columns), device=self.dev)
        )

    def add_batch(self, inp: torch.Tensor, out: torch.Tensor):
        """
        Add a batch of layer input and output data to the Hessian calculation

        :param inp: tensor containing layer input
        :param out: tensor containing layer output
        """
        if len(inp.shape) == 2:
            inp = inp.unsqueeze(0)
        tmp = inp.shape[0]
        if isinstance(self.layer, nn.Linear) or isinstance(
            self.layer, transformers.Conv1D
        ):
            if len(inp.shape) == 3:
                inp = inp.reshape((-1, inp.shape[-1]))
            inp = inp.t()
        self.H *= self.nsamples / (self.nsamples + tmp)
        self.nsamples += tmp
        inp = math.sqrt(2 / self.nsamples) * inp.float()
        self.H += inp.matmul(inp.t())

    def compress(
        self,
        blocksize: int = 128,
        percdamp: float = 0.01,
    ):
        """
        Run pruning and quantization(if applicable) on the layer up to the target
        sparsity value.

        :param blocksize: Number of columns to compress in one pass
        :param percdamp: Amount of dampening to apply to H, as a fraction of the
            diagonal norm
        """
        args_loc = "quantization_scheme.weights"
        weight_quant_args = getattr_chain(self.layer, args_loc, None)
        if weight_quant_args is None:
            logger.debug(f"Skipping unquantized layer {self.name}...")
            return

        if is_module_offloaded(self.layer):
            self.layer._hf_hook.pre_forward(self.layer)

        strategy = weight_quant_args.strategy
        actorder = weight_quant_args.actorder
        final_shape = self.layer.weight.shape
        final_dtype = self.layer.weight.dtype
        W = self.layer.weight.data.clone()

        # standardize shape and dtype
        if isinstance(self.layer, nn.Conv2d):
            W = W.flatten(1)
        elif isinstance(self.layer, transformers.Conv1D):
            W.transpose_(0, 1)
        W = W.float()

        tick = time.time()

        if strategy == QuantizationStrategy.GROUP:
            # mapping from column index to group index
            g_idx = (
                torch.arange(self.columns, device=W.device, dtype=torch.int)
                // weight_quant_args.group_size
            )

            if actorder == ActivationOrdering.GROUP:
                # permute by activation order first, then update groups
                W, self.H, perm = self._apply_activation_ordering(W, self.H)
                self._update_quantization_parameters(weight_quant_args, W)

                # use identity g_idx (invert permutation later)

            elif actorder == ActivationOrdering.WEIGHT:
                # update groups first, then permute by activation order
                self._update_quantization_parameters(weight_quant_args, W)
                W, self.H, perm = self._apply_activation_ordering(W, self.H)

                # permute g_idx to maintain identity mapping after unpermutation
                g_idx = g_idx[perm]

        scale = self.layer.weight_scale
        zero_point = self.layer.weight_zero_point

        # sparsity mask
        sparsity = tensor_sparsity(W)
        preserve_zeros = sparsity >= SPARSITY_THRESHOLD
        W_nz_mask = (
            (~torch.isclose(W, torch.zeros(1, device=W.device).float())).float()
            if preserve_zeros
            else None
        )

        # mask dead hessian values
        dead = torch.diag(self.H) == 0
        self.H[dead, dead] = 1
        W[:, dead] = 0

        Losses = torch.zeros(self.rows, device=self.dev)

        # compute inverse hessian in place to save memory
        damp = percdamp * torch.mean(torch.diag(self.H))
        diag = torch.arange(self.columns, device=self.dev)
        self.H[diag, diag] += damp
        self.H = torch.linalg.cholesky(self.H)
        self.H = torch.cholesky_inverse(self.H)
        self.H = torch.linalg.cholesky(self.H, upper=True)
        Hinv = self.H

        # See section 3.4 of https://arxiv.org/abs/2203.07259
        for i1 in range(0, self.columns, blocksize):
            i2 = min(i1 + blocksize, self.columns)
            count = i2 - i1

            W1 = W[:, i1:i2].clone()
            Q1 = torch.zeros_like(W1)
            Err1 = torch.zeros_like(W1)
            Losses1 = torch.zeros_like(W1)
            Hinv1 = Hinv[i1:i2, i1:i2]

            if preserve_zeros:
                W1_nz_mask = W_nz_mask[:, i1:i2]

            for i in range(count):
                w = W1[:, i]
                d = Hinv1[i, i]
                q = w.clone()

                # quantize column
                if strategy == QuantizationStrategy.TENSOR:
                    q = fake_quantize(
                        q,
                        scale,
                        zero_point,
                        self.layer.quantization_scheme.weights,
                    )
                elif strategy == QuantizationStrategy.CHANNEL:
                    q = fake_quantize(
                        q,
                        scale[:, 0],
                        zero_point[:, 0],
                        weight_quant_args,
                    )
                elif strategy == QuantizationStrategy.GROUP:
                    # get the group index for the current column
                    column_idx = i1 + i
                    group_index = g_idx[column_idx]

                    # Since we're only applying quantization to a slice, this
                    # ends up being a channelwise application
                    altered_qargs = copy(weight_quant_args)
                    altered_qargs.strategy = QuantizationStrategy.CHANNEL
                    q = fake_quantize(
                        q,
                        scale[:, group_index],
                        zero_point[:, group_index],
                        altered_qargs,
                    )
                else:
                    raise ValueError(
                        "Quantization strategy is not supported for GPTQ: "
                        f"{strategy}"
                    )

                # propagate column error
                Q1[:, i] = q
                Losses1[:, i] = (w - q) ** 2 / d**2

                err1 = (w - q) / d
                w1_err = err1.unsqueeze(1).matmul(Hinv1[i, i:].unsqueeze(0))
                if preserve_zeros:
                    W1[:, i:] -= w1_err * W1_nz_mask[:, i:]
                else:
                    W1[:, i:] -= w1_err
                Err1[:, i] = err1

            # propagate block error
            W[:, i1:i2] = Q1
            Losses += torch.sum(Losses1, 1) / 2

            w_err = Err1.matmul(Hinv[i1:i2, i2:])
            if preserve_zeros:
                W[:, i2:] -= w_err * W_nz_mask[:, i2:]
            else:
                W[:, i2:] -= w_err

        if "METRIC" in logger._core.levels.keys():
            self._log_metrics(tick, Losses)

        if strategy == QuantizationStrategy.GROUP:
            if actorder == ActivationOrdering.WEIGHT:
                # restore original permutation
                invperm = torch.argsort(perm)
                W = W[:, invperm]

            elif actorder == ActivationOrdering.GROUP:
                # restore original permutation
                invperm = torch.argsort(perm)
                W = W[:, invperm]
                g_idx = g_idx[invperm]

                # only save g_idx if mapping is not identity
                update_parameter_data(self.layer, g_idx, "weight_g_idx")

        if isinstance(self.layer, transformers.Conv1D):
            W.transpose_(0, 1)
        W = W.reshape(final_shape).to(final_dtype)

        # This is a bit hacky, but FSDP updates only work if we change
        # the weight in place, clone() or direct assignment won't work
        self.layer.weight -= self.layer.weight
        self.layer.weight += W

        if is_module_offloaded(self.layer):
            device = get_offloaded_device(self.layer)
            update_prefix_dict(self.layer, "weight", self.layer.weight.to(device))
            self.layer._hf_hook.post_forward(self.layer, None)

    def free(self):
        """
        Free the Hessian memory after the layer is complete
        """
        delattr(self, "H")
        super().free()

    def _update_quantization_parameters(self, args: QuantizationArgs, W: torch.Tensor):
        """
        Update layer quantization parameters with potentially permuted weight

        :param args: quantization arguments
        :param W: weight to calculate quantization parameters from
        """
        observer = args.get_observer()
        _scale, _zero_point = observer(W, g_idx=None)
        update_parameter_data(self.layer, _scale, "weight_scale")
        update_parameter_data(self.layer, _zero_point, "weight_zero_point")

    def _apply_activation_ordering(
        self, W: torch.Tensor, H: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Permute weight and hessian in order of greatest outupt activations

        :param W: weight to permute
        """
        perm = torch.argsort(torch.diag(H), descending=True)
        return W[:, perm], H[perm][:, perm], perm

    def _log_metrics(self, start_tick: float, losses: torch.Tensor):
        """
        Log metrics related to compression algorithm

        :param start_tick: time when algorithm started"
        :param losses: loss as result of algorithm
        """
        patch = logger.patch(lambda r: r.update(function="compress"))
        patch.log("METRIC", "time %.2f" % (time.time() - start_tick))
        patch.log("METRIC", "error %.2f" % torch.sum(losses).item())

        gpu_usage = get_GPU_memory_usage()
        if len(gpu_usage) > 0:
            for i in range(len(gpu_usage)):
                perc = gpu_usage[i][0] * 100
                total_memory = int(gpu_usage[i][1])  # GB
                patch.log(
                    "METRIC",
                    (
                        f"GPU {i} | usage: {perc:.2f}%"
                        f" | total memory: {total_memory} GB"
                    ),
                )

        patch.log(
            "METRIC",
            f"Compressed layer size: {get_layer_size_bytes(self.layer)} MB",
        )



import operator
from typing import Dict, Tuple

import torch
from compressed_tensors import get_execution_device
from loguru import logger
from torch.nn import Module
from tqdm import tqdm

from llmcompressor.modifiers.utils.compression_wrapper import ModuleCompressionWrapper
from llmcompressor.modifiers.utils.pytorch_helpers import EarlyStopException
from llmcompressor.pytorch.utils import tensors_to_device
from llmcompressor.utils.fsdp.context import (
    fix_fsdp_module_name,
    summon_full_params_context,
)
from llmcompressor.utils.pytorch import set_layer
from llmcompressor.utils.pytorch.module import get_prunable_layers

__all__ = ["LayerCompressor"]


class LayerCompressor:
    """
    Runs weight sparisification on a single layer using calibration data inputs. The
    layer may contain submodules. The specific sparsification algorithm is determined
    by module_compressor_class.

    Lifecycle:
        - pre_compress()
            - compressible_modules()
            - module_compressor_class.register_forward_hook()
        - compress()
            - module_compressor_class.compress()
        - post_compress()
        - revert_layer_wrappers()

    :param module_compressor_class: wrapper class to use for root modules
    :param model: model containing the layer we are running compression on
    :param layer: layer to run compression on
    :param layer_index: index of layer in the model
    :param args: additional keyword arguments
    """

    def __init__(
        self,
        module_compressor_class: ModuleCompressionWrapper,
        model: Module,
        layer: Module,
        layer_index: int,
        name: str,
        args: Dict,
    ):
        self.module_compressor_class = module_compressor_class
        self.model = model
        self.layer = layer
        self.layer_index = layer_index
        self.name = name
        self.args = args
        self.handles = None
        self.early_stop_handle = None
        self.modules = {}

    def compressible_modules(self) -> Dict:
        """
        Get the list of modules in the layer that can be compressed

        :return: dictionary of compressible modules
        """
        compressible_layers = get_prunable_layers(self.layer)
        return compressible_layers

    def set_early_stop(self):
        """
        Adds an early stopping exception to the input of the layer. This will cause the
        model to immediately exit the forward pass when reaching this layer.
        """

        def trigger_early_stop_fn(self, args, kwargs):
            raise EarlyStopException(args, kwargs)

        self.early_stop_handle = self.layer.register_forward_pre_hook(
            trigger_early_stop_fn, with_kwargs=True
        )

    def clear_early_stop(self):
        """
        Clears the early stopping handle
        """
        if self.early_stop_handle is not None:
            self.early_stop_handle.remove()
            self.early_stop_handle = None

    def pre_compress(self):
        """
        Sets up the CompressionWrapper objects for each compressible module, adding a
        hook for computing the Hessians as calibration data is passed through.
        """
        subset = self.compressible_modules()

        for name in subset:
            layer = subset[name]
            full_name = self._get_full_submodule_name(name)
            with summon_full_params_context(self.layer):
                wrapper = self.module_compressor_class(full_name, layer)
            if len(name) == 0:  # special case if layer has no children (i.e. lm_head)
                with summon_full_params_context(self.model):
                    set_layer(full_name, wrapper, self.model)
            else:
                set_layer(name, wrapper, self.layer)
            self.modules[name] = wrapper

        self.layer = operator.attrgetter(self.name)(self.model)

        def add_batch(name):
            def tmp(_, inp, out):
                self.modules[name].add_batch(inp[0].data, out.data)

            return tmp

        self.handles = []
        for name in self.modules:
            self.handles.append(subset[name].register_forward_hook(add_batch(name)))

    def calibrate_layer(self, intermediates: Tuple[Tuple, Dict]) -> Tuple[Tuple, Dict]:
        """
        Runs all calibration samples through the stored layer

        :param intermediates: inputs to run through the layer
        :return: outputs of the layer
        """
        for idx in tqdm(range(len(intermediates))):
            args, kwargs = intermediates[idx]
            device = get_execution_device(self.layer)
            output = self.layer(*tensors_to_device(args, device), **kwargs)
            intermediates[idx] = (tensors_to_device(output, "cpu"), kwargs)
            torch.cuda.empty_cache()

        return intermediates

    def post_compress(self):
        """
        remove the add_batch forward hooks after compression is complete
        """
        for handle in self.handles:
            handle.remove()

    def revert_layer_wrappers(self):
        """
        Reverts wrapped root modules back to their original structure
        """
        for name, module_wrapper in self.modules.items():
            full_name = self._get_full_submodule_name(name)
            if len(name) == 0:  # special case if layer has no children (i.e. lm_head)
                with summon_full_params_context(self.model):
                    set_layer(full_name, module_wrapper.layer, self.model)
            else:
                set_layer(name, module_wrapper.layer, self.layer)
            torch.cuda.empty_cache()
        self.modules = None

    def compress(self):
        """
        Apply compression to each wrapped submodule in the layer
        """

        @torch.no_grad()
        def compress_module(module):
            if isinstance(module, self.module_compressor_class):
                full_name = self._get_full_submodule_name(module.name)
                logger.info(f"Compressing {full_name}...")
                module.compress(**self.args)
                module.free()
                print("done")

        self.layer.apply(compress_module)
        torch.cuda.empty_cache()

    def _get_full_submodule_name(self, name):
        full_name = ".".join(x for x in [self.name, name] if len(x) > 0)
        full_name = fix_fsdp_module_name(full_name)
        return full_name



# flake8: noqa

from .data import DataTrainingArguments, TextGenerationDataset
from .model_args import ModelArguments
from .session_mixin import SessionManagerMixIn
from .text_generation import apply, compress, eval, oneshot, train
from .training_args import TrainingArguments



"""
Module for generating and managing version information for LLM Compressor.

This module provides functionality for creating semantic version strings based on the
version base and build type. It supports `release`, `nightly`, and `dev` build types.
"""

from typing import Optional, Tuple

# Define the base version and build type
version_base = "0.2.0"
build_type = "nightly"  # can be 'release', 'nightly', 'dev', or 'dev' with a dev number


def _generate_version_attributes(
    base, type_
) -> Tuple[str, int, int, int, Optional[str]]:
    from datetime import datetime

    parts = base.split(".")
    major, minor, patch = int(parts[0]), int(parts[1]), int(parts[2])

    if type_ == "release":
        build = None
    elif type_ == "nightly":
        build = datetime.utcnow().strftime("%Y%m%d")
    elif "dev" in type_:
        build = type_
    else:
        raise ValueError(f"Unknown build type: {type_}")

    ver = f"{major}.{minor}.{patch}"
    if build:
        ver += f".{build}"

    return ver, major, minor, patch, build


version, version_major, version_minor, version_patch, version_build = (
    _generate_version_attributes(version_base, build_type)
)
__version__ = version


__all__ = [
    "__version__",
    "version_base",
    "build_type",
    "version",
    "version_major",
    "version_minor",
    "version_patch",
    "version_build",
]
